import pytest

from common.version import InvalidVersionError, GenericVersion

from indy_common.version import NodeVersion
from indy_node.utils.node_control_utils import DebianVersion, NodeControlUtil, ShellError

# TODO
# - conditionally skip all tests for non-debian systems

generated_command = None


class UpstreamTest(NodeVersion):
    pass


class DebianVersionTest(DebianVersion):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, upstream_cls=UpstreamTest, **kwargs)


@pytest.fixture(autouse=True)
def clear_cache():
    DebianVersionTest.clear_cache()


@pytest.fixture
def catch_generated_command(monkeypatch):
    global generated_command
    generated_command = None

    def _f(command, *args, **kwargs):
        global generated_command
        generated_command = command
        return ''

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    monkeypatch.setattr(NodeControlUtil, 'run_shell_command', _f)


# TODO coverage
@pytest.mark.parametrize(
    'version',
    [
        '',
        '1:'
        '-1'
    ]
)
def test_invalid_no_upstream(version):
    with pytest.raises(InvalidVersionError):
        DebianVersionTest(version)


@pytest.mark.parametrize(
    'version',
    [
        'a:1.2.3',
        '12:3.4',
        '1.:2.3',
    ]
)
def test_invalid_epoch(version):
    with pytest.raises(InvalidVersionError):
        DebianVersionTest(version)


@pytest.mark.parametrize(
    'version',
    [
        'a1.2.3',
        '1.2.3:3',
        # '1.2', TODO uncomment once new release logic becomes completed
    ]
)
def test_invalid_upstream(version):
    with pytest.raises(InvalidVersionError):
        DebianVersionTest(version)


def test_invalid_upstream_keep_tilde():
    with pytest.raises(InvalidVersionError):
        DebianVersionTest('1.2.3~rc1', keep_tilde=True)


# TODO coverage
@pytest.mark.parametrize(
    'epoch,upstream,revision',
    [
        ('1', '1.2.3', ''),
        ('', '1.2.3~rc4', ''),
        ('', '1.2.3~rc4', '5'),
        ('', '1.2.3~dev4', '5'),
    ]
)
def test_valid_version(epoch, upstream, revision):
    version = "{}{}{}".format(
        epoch + ':' if epoch else '',
        upstream,
        '-' + revision if revision else ''
    )

    dv = DebianVersionTest(version)
    epoch = epoch if epoch else None
    upstream = UpstreamTest(upstream.replace('~', '.'))
    revision = revision if revision else None

    assert dv.epoch == epoch
    assert dv.upstream == upstream
    assert type(dv.upstream) == UpstreamTest
    assert dv.revision == revision
    assert dv.full == version
    assert dv.parts == (epoch, upstream, revision)
    assert dv.release == version
    assert dv.release_parts == (epoch, upstream, revision)


def test_default_upstream_cls():
    assert isinstance(DebianVersion('1.2.3').upstream, GenericVersion)


def test_generated_cmd_cmp(catch_generated_command):
    version1 = '1.2.2'
    version2 = '1.2.3'
    DebianVersionTest(version1) == DebianVersionTest(version2)
    assert generated_command == (
        "dpkg --compare-versions {} gt {}".format(version1, version2)
    )


def test_equal_no_shell_cmd(monkeypatch):
    called = 0
    run_shell_script_extended = NodeControlUtil.run_shell_script_extended

    def _f(*args, **kwargs):
        nonlocal called
        called += 1
        return run_shell_script_extended(*args, **kwargs)

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    assert DebianVersionTest('1.2.3') == DebianVersionTest('1.2.3')
    assert not called


def test_comparison_operators():
    assert DebianVersionTest('1.2.2') < DebianVersionTest('1.2.3')
    assert DebianVersionTest('1.2.3') > DebianVersionTest('1.2.2')
    assert DebianVersionTest('1.2.3') == DebianVersionTest('1.2.3')
    assert DebianVersionTest('1.2.2') <= DebianVersionTest('1.2.3')
    assert DebianVersionTest('1.2.3') >= DebianVersionTest('1.2.2')
    assert DebianVersionTest('1.2.3') != DebianVersionTest('1.2.2')


def test_compare_called_once(monkeypatch):
    called = 0
    run_shell_script_extended = NodeControlUtil.run_shell_script_extended

    def _f(*args, **kwargs):
        nonlocal called
        called += 1
        return run_shell_script_extended(*args, **kwargs)

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    assert DebianVersionTest('1.2.3') > DebianVersionTest('1.2.2')
    assert DebianVersionTest('1.2.3') > DebianVersionTest('1.2.2')
    assert DebianVersionTest('1.2.2') < DebianVersionTest('1.2.3')
    assert called == 1


@pytest.mark.parametrize(
    'v1,v2,expected',
    [
        ('1.2.3', '1.2.2', True),
        ('1.2.3~dev1', '1.2.3', False),
        ('1.2.3~dev2', '1.2.3~dev1', True),
        ('1.2.3~rc1', '1.2.3', False),
        ('1.2.3~rc1', '1.2.3~dev1', True),
        ('1.2.3~rc2', '1.2.3~rc1', True),
    ]
)
def test_compare_valid(v1, v2, expected):
    res = DebianVersionTest(v1) > DebianVersionTest(v2)
    assert res if expected else not res


def test_compare_shell_error(monkeypatch):
    run_shell_script_extended = NodeControlUtil.run_shell_script_extended

    def _f(command, *args, **kwargs):
        return run_shell_script_extended('unknown command run', *args, **kwargs)

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    with pytest.raises(ShellError):
        DebianVersionTest('1.2.3') > DebianVersionTest('1.2.2')



from indy_common.authorize.auth_request_validator import WriteRequestValidator

from indy_common.constants import RICH_SCHEMA

from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler

from plenum.server.database_manager import DatabaseManager
from stp_core.common.log import getlogger

logger = getlogger()


class RichSchemaHandler(AbstractRichSchemaObjectHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(RICH_SCHEMA, database_manager, write_req_validator)

    def is_json_ld_content(self):
        return True

    def do_static_validation_content(self, content_as_dict, request):
        pass

    def do_dynamic_validation_content(self, request):
        pass



from indy_node.utils.node_control_tool import NodeControlTool
from plenum.test.helper import randomText
from indy_node.utils.node_control_utils import NodeControlUtil, MAX_DEPS_DEPTH


def testNodeControlResolvesDependencies(monkeypatch, tconf):
    nct = NodeControlTool(config=tconf)
    node_package = ('indy-node', '0.0.1')
    plenum_package = ('indy-plenum', '0.0.3')
    node_package_with_version = '{}={}'.format(*node_package)
    plenum_package_with_version = '{}={}'.format(*plenum_package)
    mock_info = {node_package_with_version: '{}\nVersion: {}\nDepends:{} (= {})\n'.format(
        randomText(100), node_package[1], *plenum_package),
        plenum_package_with_version: '{}'.format(randomText(100))
    }

    def mock_get_info_from_package_manager(*package):
        ret = ""
        for p in package:
            ret += mock_info.get(p, "")
        return ret

    monkeypatch.setattr(NodeControlUtil, 'update_package_cache', lambda *x: None)
    monkeypatch.setattr(NodeControlUtil, 'get_sys_holds',
                        lambda *x: [node_package[0], plenum_package[0]])
    monkeypatch.setattr(NodeControlUtil, '_get_info_from_package_manager',
                        lambda *x: mock_get_info_from_package_manager(*x))
    ret = nct._get_deps_list(node_package_with_version)
    nct.server.close()
    assert sorted(ret.split()) == sorted([plenum_package_with_version,
                                          node_package_with_version])


def test_create_deps_for_exotic_version_style():
    depends = ['package1', 'package2']
    versions = ['1.6.74', '0.9.4~+.-AbCd1.2.3.4.EiF']
    def mock_info_from_package_manager(*package):
        pkg_info = """Package: {package}
Version: 1.1.26
Priority: extra
Section: default
Maintainer: Some Organization <some_org@org.com>
Installed-Size: 21.5 kB
Depends: {dep1} (= {ver1}), {dep2} (= {ver2})
Homepage: https://github.com/some/package
License: Apache 2.0
Vendor: none
Download-Size: 10.4 kB
APT-Sources: https://some.org/deb xenial/rc amd64 Packages
Description: Some package
""".format(**{'package': package[0],
            'dep1': depends[0],
            'dep2': depends[1],
            'ver1': versions[0],
            'ver2': versions[1]})
        return pkg_info
    ncu = NodeControlUtil
    ncu._get_info_from_package_manager = mock_info_from_package_manager
    ret = ncu.get_deps_tree('package', depth=MAX_DEPS_DEPTH-1)
    """
    Expected return value is:
    0 item is package,
    1 deps with version for previous package,
    """
    assert len(ret[1][0]) == len(depends)
    assert any([depends[0] in l for l in ret[1][0]])
    assert any([depends[0] in l for l in ret[1][0]])
    assert any(["{}={}".format(depends[0], versions[0]) in l for l in ret[1][0]])
    assert any(["{}={}".format(depends[1], versions[1]) in l for l in ret[1][0]])


def test_max_depth_for_deps_tree():
    depends = ['package1', 'package2']
    def mock_info_from_package_manager(*package):
        pkg_info = """Depends: {} (= 1.1.1), {} (= 2.2.2)""".format(depends[0],
                                                                    depends[1])
        return pkg_info

    ncu = NodeControlUtil
    ncu._get_info_from_package_manager = mock_info_from_package_manager
    ret = ncu.get_deps_tree('package')
    assert len(ret) <= MAX_DEPS_DEPTH



from binascii import hexlify
import pytest

from indy.ledger import build_get_nym_request
from indy_common.constants import ENDORSER_STRING
from indy_node.test.helper import check_str_is_base58_compatible, modify_field, \
    createUuidIdentifier, createHalfKeyIdentifierAndAbbrevVerkey, createCryptonym
from indy_node.test.nym_txn.test_nym_additional import get_nym

from plenum.test.helper import sdk_get_and_check_replies
from plenum.common.util import friendlyToRaw
from plenum.common.exceptions import RequestNackedException
from plenum.common.constants import IDENTIFIER
from plenum.test.pool_transactions.helper import sdk_add_new_nym, sdk_sign_and_send_prepared_request


def testSendGetNymSucceedsForExistingUuidDest(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    new_wallet = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, new_wallet[1])


def testSendGetNymFailsForNotExistingUuidDest(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, createUuidIdentifier())


def test_get_nym_returns_role(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    current_role = ENDORSER_STRING
    uuidIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                    dest=uuidIdentifier, verkey=abbrevVerkey, role=current_role)
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, createUuidIdentifier())

    new_role = ''
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                    dest=uuidIdentifier, role=new_role, skipverkey=True)
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, createUuidIdentifier())


def testSendGetNymFailsIfCryptonymIsPassedAsDest(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    get_nym(looper, sdk_pool_handle, sdk_wallet_trustee, createCryptonym())


def testSendGetNymFailsIfDestIsPassedInHexFormat(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    # Sometimes hex representation can use only base58 compatible characters
    while True:
        uuidIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
        hexEncodedUuidIdentifier = hexlify(
            friendlyToRaw(uuidIdentifier)).decode()
        if not check_str_is_base58_compatible(hexEncodedUuidIdentifier):
            break
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee, dest=uuidIdentifier, verkey=abbrevVerkey)

    _, s_did = sdk_wallet_trustee
    get_nym_req = looper.loop.run_until_complete(build_get_nym_request(s_did, uuidIdentifier))
    get_nym_req = modify_field(get_nym_req, hexEncodedUuidIdentifier, IDENTIFIER)
    req = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                             sdk_pool_handle, get_nym_req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [req])
    e.match('should not contain the following chars')


def testSendGetNymFailsIfDestIsInvalid(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    uuidIdentifier = createUuidIdentifier()
    invalidIdentifier = uuidIdentifier[:-4]
    _, s_did = sdk_wallet_trustee
    get_nym_req = looper.loop.run_until_complete(build_get_nym_request(s_did, uuidIdentifier))
    get_nym_req = modify_field(get_nym_req, invalidIdentifier, IDENTIFIER)
    req = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                             sdk_pool_handle, get_nym_req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [req])
    e.match('should be one of \[16, 32\]')


def testSendGetNymHasInvalidSyntaxIfDestIsEmpty(
        looper, sdk_pool_handle, sdk_wallet_trustee):
    uuidIdentifier = createUuidIdentifier()
    _, s_did = sdk_wallet_trustee
    get_nym_req = looper.loop.run_until_complete(build_get_nym_request(s_did, uuidIdentifier))
    get_nym_req = modify_field(get_nym_req, '', IDENTIFIER)
    req = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                             sdk_pool_handle, get_nym_req)
    with pytest.raises(RequestNackedException) as e:
        sdk_get_and_check_replies(looper, [req])
    e.match('client request invalid')



from indy_common.constants import SCHEMA_NAME, SCHEMA_VERSION, GET_SCHEMA
from indy_common.req_utils import get_read_schema_from, get_read_schema_name, get_read_schema_version
from indy_node.server.request_handlers.domain_req_handlers.schema_handler import SchemaHandler
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler


class GetSchemaHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, GET_SCHEMA, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        author_did = get_read_schema_from(request)
        schema_name = get_read_schema_name(request)
        schema_version = get_read_schema_version(request)
        schema, last_seq_no, last_update_time, proof = self.get_schema(
            author=author_did,
            schema_name=schema_name,
            schema_version=schema_version,
            with_proof=True
        )
        # TODO: we have to do this since SCHEMA has a bit different format than other txns
        # (it has NAME and VERSION inside DATA, and it's not part of the state value, but state key)
        if schema is None:
            schema = {}
        schema.update({
            SCHEMA_NAME: schema_name,
            SCHEMA_VERSION: schema_version
        })
        return self.make_result(request=request,
                                data=schema,
                                last_seq_no=last_seq_no,
                                update_time=last_update_time,
                                proof=proof)

    def get_schema(self,
                   author: str,
                   schema_name: str,
                   schema_version: str,
                   is_committed=True,
                   with_proof=True) -> (str, int, int, list):
        assert author is not None
        assert schema_name is not None
        assert schema_version is not None
        path = SchemaHandler.make_state_path_for_schema(author, schema_name, schema_version)
        try:
            keys, seq_no, last_update_time, proof = self.lookup(path, is_committed, with_proof=with_proof)
            return keys, seq_no, last_update_time, proof
        except KeyError:
            return None, None, None, None



from plenum.common.ledger import Ledger
from plenum.common.exceptions import UnknownIdentifier
from plenum.common.constants import TARGET_NYM, VERKEY, IDENTIFIER
from plenum.common.txn_util import get_payload_data
from plenum.server.client_authn import NaclAuthNr


class NodeAuthNr(NaclAuthNr):
    def __init__(self, ledger: Ledger):
        self.ledger = ledger

    def getVerkey(self, ident, request):
        # TODO: This is very inefficient
        verkey = None
        found = False
        for _, txn in self.ledger.getAllTxn():
            txn_data = get_payload_data(txn)
            if txn_data[TARGET_NYM] == ident:
                found = True
                if txn_data.get(VERKEY):
                    verkey = txn_data[VERKEY]

        if not found:
            raise UnknownIdentifier(ident)
        verkey = verkey or ident
        return verkey



import pytest

from plenum.test.pool_transactions.helper import sdk_add_new_nym


@pytest.fixture(scope="module")
def some_transactions_done(looper, nodeSet, sdk_pool_handle,
                           sdk_wallet_trustee):
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)
    # TODO: Since empty verkey and absence of verkey are stored in the ledger
    # in the same manner, this fails during catchup since the nodes that
    # processed the transaction saw verkey as `''` but while deserialising the
    # ledger they cannot differentiate between None and empty string.
    # updateIndyIdrWithVerkey(looper, new_w, new_c,
    #                           new_idr, '')



from abc import ABCMeta, abstractmethod
from collections import deque
from datetime import timedelta
from typing import Callable

from stp_core.common.log import getlogger
from plenum.server.has_action_queue import HasActionQueue
from plenum.server import notifier_plugin_manager
import asyncio

logger = getlogger()


class NodeMaintainer(HasActionQueue, metaclass=ABCMeta):
    defaultActionTimeout = 10  # minutes

    @staticmethod
    def get_timeout(timeout):
        return timedelta(minutes=timeout).seconds

    def __init__(self,
                 nodeId,
                 nodeName,
                 dataDir,
                 config,
                 ledger=None,
                 actionLog=None,
                 actionFailedCallback: Callable = None,
                 action_start_callback: Callable = None):

        self.nodeId = nodeId
        self.nodeName = nodeName
        self.config = config
        self.dataDir = dataDir
        self.ledger = ledger
        self.scheduledAction = None
        self._notifier = notifier_plugin_manager.PluginManager()
        self._actionLog = actionLog if actionLog else \
            self._defaultLog(dataDir, config)
        self._actionFailedCallback = \
            actionFailedCallback if actionFailedCallback else lambda: None
        self._action_start_callback = \
            action_start_callback if action_start_callback else lambda: None

        self.retry_timeout = 5
        self.retry_limit = 3

        self.process_action_log_for_first_run()

        HasActionQueue.__init__(self)

    def __repr__(self):
        # Since nodeid can be null till pool ledger has not caught up
        return self.nodeId or ''

    def service(self):
        return self._serviceActions()

    def process_action_log_for_first_run(self):
        # whether action was started before the Node restarted,
        # that is whether Action Log contains STARTED event
        if self._is_action_started():
            # append SUCCESS to the action log
            self._update_action_log_for_started_action()

    @property
    def lastActionEventInfo(self):
        """
        action parameters of last performed action

        :returns: action parameters or None if there were no actions
        """
        return self._actionLog.last_event

    def _unscheduleAction(self):
        """
        Unschedule current action

        Note that it does not add record to action log and does not do
        required steps to resume previous action. If you need this - use
        _cancelScheduledAction

        """
        logger.trace("{} unscheduling actions".format(self))
        self.aqStash = deque()
        self.scheduledAction = None

    async def _open_connection_and_send(self, message: str):
        controlServiceHost = self.config.controlServiceHost
        controlServicePort = self.config.controlServicePort
        msgBytes = bytes(message, "utf-8")
        _, writer = await asyncio.open_connection(
            host=controlServiceHost,
            port=controlServicePort
        )
        writer.write(msgBytes)
        writer.close()

    @abstractmethod
    def _defaultLog(self, dataDir, config):
        """
        Default log for store action txns
        :param dataDir:
        :param config:
        :return: ActionLog
        """

    @abstractmethod
    def _update_action_log_for_started_action(self):
        pass

    @abstractmethod
    def _is_action_started(self):
        pass


class NodeControlToolMessage(metaclass=ABCMeta):
    """
    Data structure that represents request for node control tool
    """

    def __init__(self, message_type: str):
        self.message_type = message_type

    def toJson(self):
        import json
        return json.dumps(self.__dict__)



import pytest

from indy_common.constants import NYM, CLAIM_DEF, REF, CLAIM_DEF_PUBLIC_KEYS, CLAIM_DEF_SCHEMA_REF
from indy_node.server.request_handlers.domain_req_handlers.claim_def_handler import ClaimDefHandler
from indy_node.test.request_handlers.helper import add_to_idr

from plenum.common.constants import STEWARD
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import reqToTxn, get_seq_no, append_txn_metadata

from plenum.common.util import randomString
from plenum.test.testing_utils import FakeSomething


@pytest.fixture(scope="module")
def claim_def_handler(db_manager, write_auth_req_validator):
    return ClaimDefHandler(db_manager, write_auth_req_validator)


@pytest.fixture(scope="module")
def creator(db_manager):
    identifier = randomString()
    idr = db_manager.idr_cache
    add_to_idr(idr, identifier, STEWARD)
    return identifier


@pytest.fixture(scope="module")
def domain_ledger():
    ledger = FakeSomething()
    ledger.txn_list = {}
    ledger.getBySeqNo = lambda seq_no: ledger.txn_list[seq_no]
    ledger.appendTxns = lambda txns: ledger.txn_list.update({get_seq_no(txn): txn
                                                             for txn in txns})
    ledger.get_by_seq_no_uncommitted = lambda seq_no: ledger.txn_list[seq_no]
    return ledger


@pytest.fixture(scope="function")
def schema(schema_request):
    return reqToTxn(schema_request)


@pytest.fixture(scope="function")
def claim_def_request(creator, schema):
    return Request(identifier=creator,
                   reqId=5,
                   signature="sig",
                   operation={'type': CLAIM_DEF,
                              'ref': get_seq_no(schema),
                              'verkey': randomString(),
                              'data': {}})


def test_claim_def_dynamic_validation_without_schema(claim_def_request,
                                                     claim_def_handler: ClaimDefHandler):
    with pytest.raises(InvalidClientRequest) as e:
        claim_def_handler.dynamic_validation(claim_def_request, 0)
    assert "Mentioned seqNo ({}) doesn't exist.".format(claim_def_request.operation[REF]) \
           in e._excinfo[1].reason


def test_claim_def_dynamic_validation_for_new_claim_def(claim_def_request, schema,
                                                        claim_def_handler: ClaimDefHandler):
    claim_def_handler.ledger.appendTxns([schema])
    claim_def_handler.dynamic_validation(claim_def_request, 0)


def test_claim_def_dynamic_validation_without_permission(claim_def_request, schema,
                                                         claim_def_handler: ClaimDefHandler):
    claim_def_handler.ledger.appendTxns([schema])
    test_identifier = randomString()
    idr = claim_def_handler.database_manager.idr_cache
    add_to_idr(idr, test_identifier, "")

    request = Request(identifier=test_identifier,
                      reqId=claim_def_request.reqId,
                      signature="sig",
                      operation=claim_def_request.operation)
    with pytest.raises(UnauthorizedClientRequest,
                       match="Not enough .* signatures"):
        claim_def_handler.dynamic_validation(request, 0)


def test_claim_def_dynamic_validation_for_unknown_identifier(claim_def_request, schema,
                                                             claim_def_handler: ClaimDefHandler):
    test_identifier = randomString()
    claim_def_handler.ledger.appendTxns([schema])
    request = Request(identifier=test_identifier,
                      reqId=claim_def_request.reqId,
                      operation=claim_def_request.operation)
    with pytest.raises(UnauthorizedClientRequest,
                       match='DID {} is not found in the Ledger'.format(test_identifier)):
        claim_def_handler.dynamic_validation(request, 0)


def test_claim_def_dynamic_validation_without_ref_to_not_schema(claim_def_request, schema,
                                                                claim_def_handler: ClaimDefHandler, creator):
    claim_def_handler.ledger.appendTxns([schema])
    nym = reqToTxn(Request(identifier=creator, operation={'type': NYM}))
    claim_def_request.operation[REF] = get_seq_no(nym)
    claim_def_handler.ledger.appendTxns([nym])
    with pytest.raises(InvalidClientRequest) as e:
        claim_def_handler.dynamic_validation(claim_def_request, 0)
    assert "Mentioned seqNo ({}) isn't seqNo of the schema.".format(claim_def_request.operation[REF]) \
           in e._excinfo[1].reason


def test_update_state(claim_def_request, claim_def_handler: ClaimDefHandler, schema_handler, schema_request):
    # add schema to state
    schema_seq_no = 1
    schema_txn_time = 1560241000
    schema_txn = reqToTxn(schema_request)
    append_txn_metadata(schema_txn, schema_seq_no, schema_txn_time)
    schema_handler.update_state(schema_txn, None, schema_request)

    # prepare claim def data
    claim_def_request.operation[REF] = schema_seq_no
    claim_def_seq_no = 1
    claim_def_txn_time = 1560241033
    claim_def_txn = reqToTxn(claim_def_request)
    append_txn_metadata(claim_def_txn, claim_def_seq_no, claim_def_txn_time)
    path, value_bytes = ClaimDefHandler.prepare_claim_def_for_state(claim_def_txn)

    claim_def_handler.update_state(claim_def_txn, None, claim_def_request)
    assert claim_def_handler.get_from_state(path) == (claim_def_request.operation[CLAIM_DEF_PUBLIC_KEYS],
                                                      claim_def_seq_no,
                                                      claim_def_txn_time)


def test_update_state_with_incorrect_data(claim_def_request, claim_def_handler: ClaimDefHandler,
                                          schema_handler, schema_request):
    # add schema to state
    schema_seq_no = 1
    schema_txn_time = 1560241000
    schema_txn = reqToTxn(schema_request)
    append_txn_metadata(schema_txn, schema_seq_no, schema_txn_time)
    schema_handler.update_state(schema_txn, None, schema_request)

    # set incorrect CLAIM_DEF_PUBLIC_KEYS
    claim_def_request.operation[CLAIM_DEF_PUBLIC_KEYS] = None
    # prepare claim def data
    claim_def_request.operation[REF] = schema_seq_no
    claim_def_seq_no = 1
    claim_def_txn_time = 1560241033
    claim_def_txn = reqToTxn(claim_def_request)
    append_txn_metadata(claim_def_txn, claim_def_seq_no, claim_def_txn_time)

    with pytest.raises(ValueError, match="'{}' field is absent, but it must contain "
                                         "components of keys".format(CLAIM_DEF_PUBLIC_KEYS)):
        claim_def_handler.update_state(claim_def_txn, None, claim_def_request)


def test_update_state_with_incorrect_schema_seq_no(claim_def_request, claim_def_handler: ClaimDefHandler,
                                                   schema_handler, schema_request):
    claim_def_seq_no = 1
    claim_def_txn_time = 1560241033
    claim_def_txn = reqToTxn(claim_def_request)
    append_txn_metadata(claim_def_txn, claim_def_seq_no, claim_def_txn_time)
    with pytest.raises(ValueError, match="'{}' field is absent, "
                                         "but it must contain schema seq no".format(CLAIM_DEF_SCHEMA_REF)):
        claim_def_handler.update_state(claim_def_txn, None, claim_def_request)



from collections import OrderedDict

from indy_common.types import ClientSchemaOperation, SchemaField, ConstraintListField, ClientAuthRuleOperation, \
    ConstraintEntityField, ConstraintField, AuthRuleValueField
from plenum.common.messages.fields import ConstantField, VersionField, IterableField, LimitedLengthStringField, \
    ChooseField, RoleField, NonNegativeNumberField, BooleanField, AnyMapField

# ConstraintListField
EXPECTED_ORDERED_FIELDS_CONSTRAINT_LIST = OrderedDict([
    ("constraint_id", ChooseField),
    ("auth_constraints", IterableField),
])


def test_has_expected_fields_list():
    actual_field_names = OrderedDict(ConstraintListField().schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS_CONSTRAINT_LIST.keys()


def test_has_expected_validators_list():
    schema = dict(ConstraintListField().schema)
    for field, validator in EXPECTED_ORDERED_FIELDS_CONSTRAINT_LIST.items():
        assert isinstance(schema[field], validator)


# ConstraintEntityField
EXPECTED_ORDERED_FIELDS_CONSTRAINT_ENTITY = OrderedDict([
    ("constraint_id", ChooseField),
    ("role", RoleField),
    ("sig_count", NonNegativeNumberField),
    ("need_to_be_owner", BooleanField),
    ("off_ledger_signature", BooleanField),
    ("metadata", AnyMapField),
])


def test_has_expected_fields_entity():
    actual_field_names = OrderedDict(ConstraintEntityField.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS_CONSTRAINT_ENTITY.keys()


def test_has_expected_validators_entity():
    schema = dict(ConstraintEntityField.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS_CONSTRAINT_ENTITY.items():
        assert isinstance(schema[field], validator)


# ClientAuthRuleOperation
EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("constraint", ConstraintField),
    ("auth_action", ChooseField),
    ("auth_type", LimitedLengthStringField),
    ("field", LimitedLengthStringField),
    ("old_value", AuthRuleValueField),
    ("new_value", AuthRuleValueField),
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientAuthRuleOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientAuthRuleOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)



import json

import pytest

from libnacl import randombytes
from plenum.common.exceptions import RequestNackedException

from plenum.common.constants import TRUSTEE, STEWARD, ROLE
from indy_common.constants import ENDORSER
from plenum.common.types import OPERATION
from plenum.common.util import randomString, hexToFriendly, friendlyToHex, rawToFriendly, friendlyToHexStr

from plenum.test.helper import sdk_get_and_check_replies, sdk_get_bad_response
from indy_node.test.helper import createUuidIdentifier, createHalfKeyIdentifierAndAbbrevVerkey, createCryptonym, \
    createUuidIdentifierAndFullVerkey
from plenum.test.pool_transactions.helper import prepare_nym_request, sdk_sign_and_send_prepared_request


@pytest.fixture(scope='module')
def nym_request(looper, sdk_wallet_trustee):
    seed = randomString(32)
    alias = randomString(5)
    dest = None
    role = None
    verkey = None
    nym_request, _ = looper.loop.run_until_complete(
        prepare_nym_request(sdk_wallet_trustee, seed,
                            alias, role, dest, verkey, True))
    return json.loads(nym_request)


def testSendNymSucceedsForUuidIdentifierAnsdk_pool_handlemittedVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': createUuidIdentifier(),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


def testSendNymSucceedsForUuidIdentifierAndFullVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': uuidIdentifier,
        'verkey': fullVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


def testSendNymSucceedsForHalfKeyIdentifierAndAbbrevVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='INDY-210')
def testSendNymFailsForCryptonymIdentifierAnsdk_pool_handlemittedVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': createCryptonym(),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='INDY-210')
def testSendNymFailsForCryptonymIdentifierAndFullVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    cryptonym = createCryptonym()

    _, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': cryptonym,
        'verkey': fullVerkey,
        'role': ENDORSER
    }

    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


def testSendNymFailsForCryptonymIdentifierAndMatchedAbbrevVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    cryptonym = createCryptonym()

    hexCryptonym = friendlyToHex(cryptonym)
    abbrevVerkey = '~' + hexToFriendly(hexCryptonym[16:])
    parameters = {
        'dest': cryptonym,
        'verkey': abbrevVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'Neither a full verkey nor an abbreviated one')


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfIdentifierSizeIs15Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(15)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException, '')


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfIdentifierSizeIs17Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(17)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfFullVerkeySizeIs31Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(16)),
        'verkey': rawToFriendly(randombytes(31)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfFullVerkeySizeIs33Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(16)),
        'verkey': rawToFriendly(randombytes(33)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfAbbrevVerkeySizeIs15Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(16)),
        'verkey': '~' + rawToFriendly(randombytes(15)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1108')
def testSendNymFailsIfAbbrevVerkeySizeIs17Bytes(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': rawToFriendly(randombytes(16)),
        'verkey': '~' + rawToFriendly(randombytes(17)),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfUuidIdentifierIsHexEncoded(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': friendlyToHexStr(createUuidIdentifier()),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfFullVerkeyIsHexEncoded(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': uuidIdentifier,
        'verkey': friendlyToHexStr(fullVerkey),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfAbbrevVerkeyIsHexEncoded(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': '~' + friendlyToHexStr(abbrevVerkey.replace('~', '')),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfIdentifierContainsNonBase58Characters(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier = createUuidIdentifier()
    parameters = {
        'dest': uuidIdentifier[:5] + '/' + uuidIdentifier[6:],
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfFullVerkeyContainsNonBase58Characters(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': uuidIdentifier,
        'verkey': fullVerkey[:5] + '/' + fullVerkey[6:],
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfAbbrevVerkeyContainsNonBase58Characters(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey[:6] + '/' + abbrevVerkey[7:],
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfFullVerkeyContainsTilde(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': uuidIdentifier,
        'verkey': '~' + fullVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1109')
def testSendNymFailsIfAbbrevVerkeysdk_pool_handleesNotContainTilde(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey.replace('~', ''),
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1110')
def testSendNymFailsIfRoleIsUnknown(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey,
        'role': 'SUPERVISOR'
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1110')
def testSendNymFailsIfRoleIsSpecifiedUsingNumericCode(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey,
        'role': ENDORSER.value
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1111')
def testSendNymHasInvalidSyntaxIfParametersOrderIsWrong(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    halfKeyIdentifier, abbrevVerkey = createHalfKeyIdentifierAndAbbrevVerkey()
    parameters = {
        'dest': halfKeyIdentifier,
        'verkey': abbrevVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1111')
def testSendNymHasInvalidSyntaxIfIdentifierIsEmpty(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    _, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': '',
        'verkey': fullVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1111')
def testSendNymHasInvalidSyntaxIfIdentifierIsOmitted(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    _, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'verkey': fullVerkey,
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


def testSendNymHasInvalidSyntaxForUuidIdentifierAndEmptyVerkey(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'dest': createUuidIdentifier(),
        'verkey': '',
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'Neither a full verkey nor an abbreviated one')


@pytest.mark.skip(reason='SOV-1111')
def testSendNymHasInvalidSyntaxIfIdentifierAndVerkeyAreOmitted(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    parameters = {
        'role': ENDORSER
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


@pytest.mark.skip(reason='SOV-1111')
def testSendNymHasInvalidSyntaxIfUnknownParameterIsPassed(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    uuidIdentifier, fullVerkey = createUuidIdentifierAndFullVerkey()
    parameters = {
        'dest': uuidIdentifier,
        'verkey': fullVerkey,
        'role': ENDORSER,
        'extra': 42
    }
    nym_request[OPERATION].update(parameters)
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_and_check_replies(looper, [request_couple])


def testSendNymHasInvalidSyntaxIfAllParametersAreOmitted(
        looper, sdk_pool_handle, txnPoolNodeSet, nym_request, sdk_wallet_trustee):
    for f in nym_request[OPERATION].keys():
        nym_request[OPERATION][f] = ''

    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                                        sdk_pool_handle, json.dumps(nym_request))
    sdk_get_bad_response(looper, [request_couple], RequestNackedException,
                         'Reason: client request invalid')



from indy_common.constants import GET_NYM, VERSION_ID, VERSION_TIME

from common.serializers.serialization import domain_state_serializer
from indy_node.server.request_handlers.domain_req_handlers.nym_handler import NymHandler
from plenum.common.constants import TARGET_NYM, TXN_TIME, DOMAIN_LEDGER_ID
from plenum.common.request import Request
from plenum.common.types import f
from plenum.common.exceptions import InvalidClientRequest
from plenum.server.database_manager import DatabaseManager

from .version_read_request_handler import VersionReadRequestHandler


class GetNymHandler(VersionReadRequestHandler):
    def __init__(self, node, database_manager: DatabaseManager):
        super().__init__(node, database_manager, GET_NYM, DOMAIN_LEDGER_ID)

    def get_result(self, request: Request):
        self._validate_request_type(request)
        nym = request.operation[TARGET_NYM]
        path = NymHandler.make_state_path_for_nym(nym)

        timestamp = request.operation.get(VERSION_TIME)
        seq_no = request.operation.get(VERSION_ID)

        if timestamp and seq_no:
            raise InvalidClientRequest(
                request.identifier,
                request.reqId,
                f"{VERSION_ID} and {VERSION_TIME} are mutually exclusive; only one should be "
                "specified",
            )
        # The above check determines whether the request is valid
        # A similar check in VersionReadRequestHandler determines
        # whether the method is used correctly

        data = None
        last_seq_no = None
        update_time = None
        proof = None

        nym_data, proof = self.lookup_version(
            path, seq_no=seq_no, timestamp=timestamp, with_proof=True
        )

        if nym_data:
            nym_data = domain_state_serializer.deserialize(nym_data)
            nym_data[TARGET_NYM] = nym
            data = domain_state_serializer.serialize(nym_data)
            last_seq_no = nym_data[f.SEQ_NO.nm]
            update_time = nym_data[TXN_TIME]

        result = self.make_result(
            request=request,
            data=data,  # Serailized retrieved txn data
            last_seq_no=last_seq_no,  # nym_data[seqNo]
            update_time=update_time,  # nym_data[TXN_TIME]
            proof=proof,  # _get_value_from_state(..., with_proof=True)[1]
        )

        result.update(request.operation)
        return result



from plenum.test.node_catchup.helper import waitNodeDataEquality
from indy_node.test.conftest import sdk_node_theta_added
from indy_node.test.helper import TestNode
from indy_common.config_helper import NodeConfigHelper
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def test_requests_post_multiple_new_nodes(
        looper,
        nodeSet,
        tconf,
        tdir,
        sdk_pool_handle,
        sdk_wallet_trustee,
        allPluginsPath,
        some_transactions_done):
    new_nodes = []
    for node_name in ('Zeta', 'Eta'):
        new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                            nodeSet,
                                                            tdir,
                                                            tconf,
                                                            sdk_pool_handle,
                                                            sdk_wallet_trustee,
                                                            allPluginsPath,
                                                            node_config_helper_class=NodeConfigHelper,
                                                            testNodeClass=TestNode,
                                                            name=node_name)
        new_nodes.append(new_node)

    for _ in range(5):
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)

    for new_node in new_nodes:
        waitNodeDataEquality(looper, new_node, *nodeSet[:-2])

    for _ in range(5):
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee)



from collections.abc import Callable

from indy_common.constants import FROM, TO, REVOC_REG_DEF_ID, ISSUANCE_TYPE, REVOKED, ISSUED, VALUE, REVOC_TYPE, \
    ACCUM_TO, STATE_PROOF_FROM, ACCUM_FROM, GET_REVOC_REG_DELTA
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_entry_handler import RevocRegEntryHandler

from indy_node.server.request_handlers.utils import StateValue
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.read_request_handler import ReadRequestHandler
from plenum.server.request_handlers.utils import decode_state_value


class GetRevocRegDeltaHandler(ReadRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 get_revocation_strategy: Callable):
        super().__init__(database_manager, GET_REVOC_REG_DELTA, DOMAIN_LEDGER_ID)
        self.get_revocation_strategy = get_revocation_strategy

    def static_validation(self, request: Request):
        operation = request.operation
        req_ts_to = operation.get(TO, None)
        assert req_ts_to
        req_ts_from = operation.get(FROM, None)

        if req_ts_from and req_ts_from > req_ts_to:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "Timestamp FROM more then TO: {} > {}".format(req_ts_from, req_ts_to))

    def get_result(self, request: Request):
        """
        For getting reply we need:
        1. Get REVOC_REG_ENTRY by "TO" timestamp from state
        2. If FROM is given in request, then Get REVOC_REG_ENTRY by "FROM" timestamp from state
        3. Get ISSUANCE_TYPE for REVOC_REG_DEF (revoked/issued strategy)
        4. Compute issued and revoked indices by corresponding strategy
        5. Make result
           5.1 Now, if "FROM" is presented in request, then STATE_PROOF_FROM and ACCUM (revocation entry for "FROM" timestamp)
               will added into data section
           5.2 If not, then only STATE_PROOF for "TO" revocation entry will added
        :param request:
        :return: Reply
        """
        self._validate_request_type(request)
        req_ts_from = request.operation.get(FROM, None)
        req_ts_to = request.operation.get(TO)
        revoc_reg_def_id = request.operation.get(REVOC_REG_DEF_ID)
        reply = None
        """
        Get root hash for "to" timestamp
        Get REVOC_REG_ENTRY and ACCUM record for timestamp "to"
        """
        path_to_reg_entry = RevocRegEntryHandler.make_state_path_for_revoc_reg_entry(revoc_reg_def_id=revoc_reg_def_id)
        path_to_reg_entry_accum = RevocRegEntryHandler.make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id=revoc_reg_def_id)

        entry_to = self._get_reg_entry_by_timestamp(req_ts_to, path_to_reg_entry)
        accum_to = self._get_reg_entry_accum_by_timestamp(req_ts_to, path_to_reg_entry_accum)
        entry_from = StateValue()
        accum_from = StateValue()

        if accum_to.value and entry_to.value:
            """Get issuance type from REVOC_REG_DEF"""
            encoded_revoc_reg_def = self.state.get_for_root_hash(entry_to.root_hash,
                                                                 revoc_reg_def_id)
            if encoded_revoc_reg_def:
                revoc_reg_def, _, _ = decode_state_value(encoded_revoc_reg_def)
                strategy_cls = self.get_revocation_strategy(revoc_reg_def[VALUE][ISSUANCE_TYPE])
                issued_to = entry_to.value[VALUE].get(ISSUED, [])
                revoked_to = entry_to.value[VALUE].get(REVOKED, [])
                if req_ts_from:
                    """Get REVOC_REG_ENTRY and ACCUM records for timestamp from if exist"""
                    entry_from = self._get_reg_entry_by_timestamp(req_ts_from, path_to_reg_entry)
                    accum_from = self._get_reg_entry_accum_by_timestamp(req_ts_from, path_to_reg_entry_accum)
                if req_ts_from and entry_from.value and accum_from.value:
                    """Compute issued/revoked lists corresponding with ISSUANCE_TYPE strategy"""
                    issued_from = entry_from.value[VALUE].get(ISSUED, [])
                    revoked_from = entry_from.value[VALUE].get(REVOKED, [])
                    result_issued, result_revoked = strategy_cls.get_delta({ISSUED: issued_to,
                                                                            REVOKED: revoked_to},
                                                                           {ISSUED: issued_from,
                                                                            REVOKED: revoked_from})
                else:
                    result_issued, result_revoked = strategy_cls.get_delta({ISSUED: issued_to,
                                                                            REVOKED: revoked_to},
                                                                           None)
                reply = {
                    REVOC_REG_DEF_ID: revoc_reg_def_id,
                    REVOC_TYPE: revoc_reg_def.get(REVOC_TYPE),
                    VALUE: {
                        ACCUM_TO: accum_to.value if entry_from.value else entry_to.value,
                        ISSUED: result_issued,
                        REVOKED: result_revoked
                    }

                }
                """If we got "from" timestamp, then add state proof into "data" section of reply"""
                if req_ts_from:
                    reply[STATE_PROOF_FROM] = accum_from.proof
                    reply[VALUE][ACCUM_FROM] = accum_from.value

        if accum_to and entry_to:
            seq_no = accum_to.seq_no if entry_from.value else entry_to.seq_no
            update_time = accum_to.update_time if entry_from.value else entry_to.update_time
            proof = accum_to.proof if entry_from.value else entry_to.proof
            if reply is None and req_ts_from is not None:
                # TODO: change this according to INDY-2115
                reply = {}
                accum_from = self._get_reg_entry_accum_by_timestamp(req_ts_from, path_to_reg_entry_accum)
                reply[STATE_PROOF_FROM] = accum_from.proof
                reply[VALUE] = {}
                reply[VALUE][ACCUM_TO] = None
                reply[VALUE][ACCUM_FROM] = accum_from.value
        else:
            seq_no = None
            update_time = None
            proof = None

        return self.make_result(request=request,
                                data=reply,
                                last_seq_no=seq_no,
                                update_time=update_time,
                                proof=proof)

    def _get_reg_entry_by_timestamp(self, timestamp, path_to_reg_entry):
        reg_entry = None
        seq_no = None
        last_update_time = None
        reg_entry_proof = None
        past_root = self.database_manager.ts_store.get_equal_or_prev(timestamp)
        if past_root:
            encoded_entry, reg_entry_proof = self._get_value_from_state(path_to_reg_entry,
                                                                        head_hash=past_root,
                                                                        with_proof=True)
            if encoded_entry:
                reg_entry, seq_no, last_update_time = decode_state_value(encoded_entry)
        return StateValue(root_hash=past_root,
                          value=reg_entry,
                          seq_no=seq_no,
                          update_time=last_update_time,
                          proof=reg_entry_proof)

    def _get_reg_entry_accum_by_timestamp(self, timestamp, path_to_reg_entry_accum):
        reg_entry_accum = None
        seq_no = None
        last_update_time = None
        reg_entry_accum_proof = None
        past_root = self.database_manager.ts_store.get_equal_or_prev(timestamp)
        if past_root:
            encoded_entry, reg_entry_accum_proof = self._get_value_from_state(
                path_to_reg_entry_accum, head_hash=past_root, with_proof=True)
            if encoded_entry:
                reg_entry_accum, seq_no, last_update_time = decode_state_value(encoded_entry)
        return StateValue(root_hash=past_root,
                          value=reg_entry_accum,
                          seq_no=seq_no,
                          update_time=last_update_time,
                          proof=reg_entry_accum_proof)



from indy_node.test import waits
from indy_node.test.upgrade.helper import bumpVersion, get_req_from_update
from plenum.common.constants import VERSION
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request
from stp_core.loop.eventually import eventually
from indy_common.version import src_version_cls


def test_forced_upgrade_no_consensus_on_single_node(
        validUpgradeExpForceTrue, looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee):
    nup = validUpgradeExpForceTrue.copy()
    nup.update({VERSION: bumpVersion(validUpgradeExpForceTrue[VERSION])})
    for node in nodeSet:
        if node.name != "Alpha" and node in looper.prodables:
            node.cleanupOnStopping = False
            looper.removeProdable(node)
            node.stop()
        else:
            node.upgrader.scheduledUpgrade = None
    _, did = sdk_wallet_trustee
    req = get_req_from_update(looper, did, nup)
    sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee, sdk_pool_handle, req)

    def testsched():
        for node in nodeSet:
            if node.name == "Alpha":
                assert node.upgrader.scheduledAction
                assert node.upgrader.scheduledAction.version.full == nup[VERSION]

    looper.run(eventually(testsched, retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))



import copy
import json

from indy_common.constants import RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE, RICH_SCHEMA_MAPPING, \
    RS_MAPPING_TYPE_VALUE, RS_CRED_DEF_TYPE_VALUE, RICH_SCHEMA_CRED_DEF, RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE, \
    JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE, RS_ID, RS_NAME, RS_TYPE, RS_VERSION, RS_CONTENT, RS_PRES_DEF_TYPE_VALUE, \
    RICH_SCHEMA_PRES_DEF
from indy_common.types import Request
from indy_node.test.rich_schema.templates import RICH_SCHEMA_EX1, W3C_BASE_CONTEXT, RICH_SCHEMA_ENCODING_EX1, \
    RICH_SCHEMA_MAPPING_EX1, RICH_SCHEMA_CRED_DEF_EX1, RICH_SCHEMA_PRES_DEF_EX1
from plenum.common.constants import TXN_TYPE, OP_VER, CURRENT_PROTOCOL_VERSION
from plenum.common.txn_util import reqToTxn, append_txn_metadata
from plenum.common.util import randomString


def rs_req(txn_type, rs_type, content, id=None):
    author = randomString()
    endorser = randomString()
    return Request(identifier=author,
                   reqId=1234,
                   signatures={author: "sig1", endorser: "sig2"},
                   endorser=endorser,
                   protocolVersion=CURRENT_PROTOCOL_VERSION,
                   operation={
                       TXN_TYPE: txn_type,
                       OP_VER: '1.1',
                       RS_ID: id or randomString(),
                       RS_NAME: randomString(),
                       RS_TYPE: rs_type,
                       RS_VERSION: '1.0',
                       RS_CONTENT: json.dumps(content)
                   })


def context_request():
    return rs_req(JSON_LD_CONTEXT, RS_CONTEXT_TYPE_VALUE,
                  content=W3C_BASE_CONTEXT)


def rich_schema_request():
    id = randomString()
    content = copy.deepcopy(RICH_SCHEMA_EX1)
    content['@id'] = id
    return rs_req(RICH_SCHEMA, RS_SCHEMA_TYPE_VALUE,
                  content=content, id=id)


def rich_schema_encoding_request():
    return rs_req(RICH_SCHEMA_ENCODING, RS_ENCODING_TYPE_VALUE,
                  content=RICH_SCHEMA_ENCODING_EX1)


def rich_schema_mapping_request():
    id = randomString()
    content = copy.deepcopy(RICH_SCHEMA_MAPPING_EX1)
    content['@id'] = id
    return rs_req(RICH_SCHEMA_MAPPING, RS_MAPPING_TYPE_VALUE,
                  content=content, id=id)


def rich_schema_cred_def_request():
    return rs_req(RICH_SCHEMA_CRED_DEF, RS_CRED_DEF_TYPE_VALUE,
                  content=RICH_SCHEMA_CRED_DEF_EX1)


def rich_schema_pres_def_request():
    id = randomString()
    content = copy.deepcopy(RICH_SCHEMA_PRES_DEF_EX1)
    content['@id'] = id
    return rs_req(RICH_SCHEMA_PRES_DEF, RS_PRES_DEF_TYPE_VALUE,
                  content=content, id=id)


def make_rich_schema_object_exist(handler, request, commit=False):
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(request)
    append_txn_metadata(txn, seq_no, txn_time)
    handler.update_state(txn, None, context_request)
    if commit:
        handler.state.commit()



import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import SCHEMA
from indy_node.test.api.helper import sdk_write_schema_and_check
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.helper import build_auth_rule_request_json


class SchemaTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()

    def run(self):
        # Step 1. Check default auth rule
        sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet, ["attrib1"], name='schema1',
                                   version='1.0')
        with pytest.raises(RequestRejectedException):
            sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet, ["attrib1"], name='schema2',
                                       version='1.0')

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 3. Check, that we cannot send schema the old way
        with pytest.raises(RequestRejectedException):
            sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet, ["attrib1"],
                                       name='schema3', version='1.0')

        # Step 4. Check, that new auth rule is used
        sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet, ["attrib1"], name='schema4',
                                   version='1.0')

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 6. Check, that default auth rule works
        sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.trustee_wallet, ["attrib1"], name='schema5',
                                   version='1.0')
        with pytest.raises(RequestRejectedException):
            sdk_write_schema_and_check(self.looper, self.sdk_pool_handle, self.new_default_wallet, ["attrib1"], name='schema6',
                                       version='1.0')

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=SCHEMA,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )



import pytest
import time

from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.authorize.authorizer import EndorserAuthorizer
from indy_common.constants import ENDORSER, NETWORK_MONITOR
from indy_common.types import Request
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import CURRENT_PROTOCOL_VERSION, STEWARD, TRUSTEE
from plenum.test.helper import randomOperation
from storage.kv_in_memory import KeyValueStorageInMemory

AUTH_CONSTR = AuthConstraint(role=ENDORSER, sig_count=1)


def get_idr_by_role_author(role):
    if role == TRUSTEE:
        return "author_did_trustee"
    if role == STEWARD:
        return "author_did_steward"
    if role == ENDORSER:
        return "author_did_endorser"
    if role == NETWORK_MONITOR:
        return "author_did_network_monitor"
    return "author_did_no_role"


def get_idr_by_role_endorser(role):
    if role == TRUSTEE:
        return "endorser_did_trustee"
    if role == STEWARD:
        return "endorser_did_steward"
    if role == ENDORSER:
        return "endorser_did_endorser"
    if role == NETWORK_MONITOR:
        return "endorser_did_network_monitor"
    return "endorser_did_no_role"


@pytest.fixture(scope='module')
def idr_cache():
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    # authors
    cache.set("author_did_no_role", 1, int(time.time()), role=IDENTITY_OWNER,
              verkey="SomeVerkey", isCommitted=False)
    cache.set("author_did_trustee", 1, int(time.time()), role=TRUSTEE,
              verkey="SomeVerkey1", isCommitted=False)
    cache.set("author_did_steward", 1, int(time.time()), role=STEWARD,
              verkey="SomeVerkey2", isCommitted=False)
    cache.set("author_did_endorser", 1, int(time.time()), role=ENDORSER,
              verkey="SomeVerkey3", isCommitted=False)
    cache.set("author_did_network_monitor", 1, int(time.time()), role=NETWORK_MONITOR,
              verkey="SomeVerkey5", isCommitted=False)

    # endorsers
    cache.set("endorser_did_no_role", 1, int(time.time()), role=IDENTITY_OWNER,
              verkey="SomeVerkey4", isCommitted=False)
    cache.set("endorser_did_trustee", 1, int(time.time()), role=TRUSTEE,
              verkey="SomeVerkey1", isCommitted=False)
    cache.set("endorser_did_steward", 1, int(time.time()), role=STEWARD,
              verkey="SomeVerkey2", isCommitted=False)
    cache.set("endorser_did_endorser", 1, int(time.time()), role=ENDORSER,
              verkey="SomeVerkey3", isCommitted=False)
    cache.set("endorser_did_network_monitor", 1, int(time.time()), role=NETWORK_MONITOR,
              verkey="SomeVerkey5", isCommitted=False)

    return cache


@pytest.fixture(scope='module')
def authorizer(idr_cache):
    return EndorserAuthorizer(cache=idr_cache)


def build_req_multi_signed_by_endorser(author_role, endorser_role=ENDORSER, append_endorser=False):
    author_idr = get_idr_by_role_author(author_role) if author_role != IDENTITY_OWNER else "author_no_role"
    endorser_idr = get_idr_by_role_endorser(endorser_role)
    req = Request(identifier=author_idr,
                  reqId=1,
                  operation=randomOperation(),
                  signatures={endorser_idr: "sig1", author_idr: "sig2"},
                  protocolVersion=CURRENT_PROTOCOL_VERSION)
    if append_endorser:
        req.endorser = endorser_idr
    return req


def build_req_multi_signed_by_author_only(author_role):
    idr = get_idr_by_role_author(author_role)
    return Request(identifier=idr,
                   reqId=1,
                   operation=randomOperation(),
                   signatures={idr: "sig2"},
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


def build_req_multi_signed_by_non_author_only(author_role, non_author_role):
    return Request(identifier=get_idr_by_role_author(author_role),
                   reqId=1,
                   operation=randomOperation(),
                   signatures={get_idr_by_role_endorser(non_author_role): "sig"},
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


def build_req_signed_by_author_only(author_role):
    idr = get_idr_by_role_author(author_role)
    return Request(identifier=idr,
                   reqId=1,
                   operation=randomOperation(),
                   signature="sig2",
                   protocolVersion=CURRENT_PROTOCOL_VERSION)


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD, ENDORSER])
def test_dont_require_endorser_if_one_sig(authorizer, author_role):
    req = build_req_signed_by_author_only(author_role)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD, ENDORSER])
def test_dont_require_endorser_if_one_multisig(authorizer, author_role):
    req = build_req_multi_signed_by_author_only(author_role)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR])
@pytest.mark.parametrize('non_author_role', [TRUSTEE, STEWARD, ENDORSER])
def test_require_endorser_if_signed_by_non_author(authorizer, author_role, non_author_role):
    req = build_req_multi_signed_by_non_author_only(author_role, non_author_role)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "'Endorser' field must be explicitly set for the endorsed transaction" in reason


@pytest.mark.parametrize('author_role', [TRUSTEE, STEWARD, ENDORSER])
@pytest.mark.parametrize('non_author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD, ENDORSER])
def test_dont_require_endorser_if_signed_by_non_author(authorizer, author_role, non_author_role):
    req = build_req_multi_signed_by_non_author_only(author_role, non_author_role)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR])
@pytest.mark.parametrize('endorser_role', [TRUSTEE, STEWARD, ENDORSER])
def test_require_endorser_when_multi_sig(authorizer, author_role, endorser_role):
    # isn't authorized without explicit Endorser field
    req = build_req_multi_signed_by_endorser(author_role, endorser_role=endorser_role, append_endorser=False)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "'Endorser' field must be explicitly set for the endorsed transaction" in reason

    # authorized with explicit Endorser field
    req = build_req_multi_signed_by_endorser(author_role, append_endorser=True)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR])
@pytest.mark.parametrize('non_author_role', [IDENTITY_OWNER, NETWORK_MONITOR])
def test_dont_require_endorser_when_multi_sig_by_owner(authorizer, author_role, non_author_role):
    req = build_req_multi_signed_by_endorser(author_role, endorser_role=non_author_role, append_endorser=False)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [TRUSTEE, STEWARD, ENDORSER])
def test_dont_require_endorser_when_multi_sig(authorizer, author_role):
    # authorized even without explicit Endorser field
    req = build_req_multi_signed_by_endorser(author_role, append_endorser=False)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized

    # authorized with explicit Endorser field
    req = build_req_multi_signed_by_endorser(author_role, append_endorser=True)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('endorser_role', [ENDORSER])
def test_allowed_endorser_roles(authorizer, endorser_role):
    req = build_req_multi_signed_by_endorser(author_role=IDENTITY_OWNER,
                                             endorser_role=endorser_role, append_endorser=True)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('endorser_role', [TRUSTEE, STEWARD, IDENTITY_OWNER, NETWORK_MONITOR])
def test_not_allowed_endorser_roles(authorizer, endorser_role):
    req = build_req_multi_signed_by_endorser(author_role=IDENTITY_OWNER,
                                             endorser_role=endorser_role, append_endorser=True)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "Endorser must have one of the following roles" in reason


@pytest.mark.parametrize('author_role', [TRUSTEE, STEWARD, ENDORSER])
def test_endorser_role_checked_when_author_is_endorser(authorizer, author_role):
    req = build_req_multi_signed_by_endorser(author_role=author_role,
                                             endorser_role=IDENTITY_OWNER, append_endorser=True)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "Endorser must have one of the following roles" in reason


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD])
def test_endorser_must_sign(authorizer, author_role):
    req = build_req_multi_signed_by_author_only(author_role=author_role)
    req.endorser = get_idr_by_role_endorser(ENDORSER)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "Endorser must sign the transaction" in reason


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD])
def test_endorser_is_author_and_1_sig(authorizer, author_role):
    req = build_req_signed_by_author_only(author_role=ENDORSER)
    req.endorser = get_idr_by_role_author(ENDORSER)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert authorized


@pytest.mark.parametrize('author_role', [IDENTITY_OWNER, NETWORK_MONITOR, TRUSTEE, STEWARD])
def test_endorser_is_not_author_and_1_sig(authorizer, author_role):
    req = build_req_signed_by_author_only(author_role=ENDORSER)
    req.endorser = get_idr_by_role_endorser(ENDORSER)
    authorized, reason = authorizer.authorize(req, AUTH_CONSTR)
    assert not authorized
    assert "Endorser must sign the transaction" in reason



import random

import pytest
from indy_common.constants import REVOC_REG_DEF, CRED_DEF_ID, REVOC_TYPE, TAG

from indy_node.persistence.idr_cache import IdrCache
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_def_handler import RevocRegDefHandler
from indy_node.server.request_handlers.domain_req_handlers.schema_handler import SchemaHandler
from indy_node.test.auth_rule.helper import generate_auth_rule_operation
from indy_node.test.request_handlers.helper import add_to_idr
from plenum.common.constants import KeyValueStorageType, TXN_TYPE, TXN_AUTHOR_AGREEMENT, TXN_AUTHOR_AGREEMENT_TEXT, \
    TXN_AUTHOR_AGREEMENT_VERSION, TXN_AUTHOR_AGREEMENT_RATIFICATION_TS
from plenum.common.request import Request
from plenum.common.util import randomString
from storage.helper import initKeyValueStorage
from indy_node.test.api.helper import req_id
_reqId = req_id()


@pytest.fixture(scope="module")
def idr_cache(tconf, tdir):
    name = 'name'
    idr_cache = IdrCache(name,
                         initKeyValueStorage(KeyValueStorageType.Rocksdb,
                                             tdir,
                                             tconf.idrCacheDbName,
                                             db_config=tconf.db_idr_cache_db_config))
    return idr_cache


@pytest.fixture(scope="module")
def schema_handler(db_manager, write_auth_req_validator):
    return SchemaHandler(db_manager, write_auth_req_validator)


@pytest.fixture(scope="function")
def schema_request():
    return Request(identifier=randomString(),
                   reqId=5,
                   signature="sig",
                   operation={'type': "101",
                              'data': {
                                  'version': '1.0',
                                  'name': 'Degree',
                                  'attr_names': ['last_name',
                                                 'first_name', ]
                              }})





@pytest.fixture(scope="function")
def rs_schema_request():
    authors_did, name, version, _type = "2hoqvcwupRTUNkXn6ArYzs", randomString(), "1.1", "8"
    _id = authors_did + ':' + _type + ':' + name + ':' + version
    return Request(identifier=authors_did,
                   reqId=next(_reqId),
                   signature="sig",
                   protocolVersion=2,
                   operation={
                       "type": "201",
                       "meta": {
                           "type": "sch",
                           "name": name,
                           "version": version
                       },
                       "data": {
                           "schema": {
                               "@id": _id,
                               "@context": "ctx:sov:2f9F8ZmxuvDqRiqqY29x6dx9oU4qwFTkPbDpWtwGbdUsrCD",
                               "@type": "rdfs:Class",
                               "rdfs:comment": "ISO18013 International Driver License",
                               "rdfs:label": "Driver License",
                               "rdfs:subClassOf": {
                                   "@id": "sch:Thing"
                               },
                               "driver": "Driver",
                               "dateOfIssue": "Date",
                               "dateOfExpiry": "Date",
                               "issuingAuthority": "Text",
                               "licenseNumber": "Text",
                               "categoriesOfVehicles": {
                                   "vehicleType": "Text",
                                   "vehicleType-input": {
                                       "@type": "sch:PropertyValueSpecification",
                                       "valuePattern": "^(A|B|C|D|BE|CE|DE|AM|A1|A2|B1|C1|D1|C1E|D1E)$"
                                   },
                                   "dateOfIssue": "Date",
                                   "dateOfExpiry": "Date",
                                   "restrictions": "Text",
                                   "restrictions-input": {
                                       "@type": "sch:PropertyValueSpecification",
                                       "valuePattern": "^([A-Z]|[1-9])$"
                                   }
                               },
                               "administrativeNumber": "Text"
                           }
                       }
                   })


@pytest.fixture(scope="function")
def rs_schema_broken_request():
    authors_did, name, version, _type = "2hoqvcwupRTUNkXn6ArYzs", randomString(), "1.1", "8"
    _id = authors_did + ':' + _type + ':' + name + ':' + version
    return Request(identifier=authors_did,
                   reqId=random.randint(1, 10000000000000000000),
                   signature="sig",
                   protocolVersion=2,
                   operation={'type': '201',
                              'data':
                                  {'schema':
                                       {'@id': '7MpRep22vL3bnyR8VqCvsS:8:ISO18023_Drivers_License:1.2',
                                        '@type': '0od'}
                                   },
                              'meta':
                                  {'type': 'sch',
                                   'version': '1.2',
                                   'name': 'ISO18023_Drivers_License'}
                              }
                   )





@pytest.fixture(scope="module")
def revoc_reg_def_handler(db_manager, write_auth_req_validator):
    return RevocRegDefHandler(db_manager, write_auth_req_validator)


@pytest.fixture(scope="module")
def revoc_reg_def_request():
    return Request(identifier=randomString(),
                   reqId=5,
                   signature="sig",
                   operation={'type': REVOC_REG_DEF,
                              CRED_DEF_ID: "credDefId",
                              REVOC_TYPE: randomString(),
                              TAG: randomString(),
                              })


@pytest.fixture(scope="module")
def creator(db_manager):
    identifier = randomString()
    idr = db_manager.idr_cache
    add_to_idr(idr, identifier, None)
    return identifier


@pytest.fixture(scope="function")
def auth_rule_request(creator):
    return Request(identifier=creator,
                   reqId=5,
                   signature="sig",
                   operation=generate_auth_rule_operation())


@pytest.fixture(scope="module")
def taa_request(creator):
    return Request(identifier=creator,
                   signature="signature",
                   operation={TXN_TYPE: TXN_AUTHOR_AGREEMENT,
                              TXN_AUTHOR_AGREEMENT_TEXT: "text",
                              TXN_AUTHOR_AGREEMENT_VERSION: "version",
                              TXN_AUTHOR_AGREEMENT_RATIFICATION_TS: 0})



import pytest

from common.exceptions import LogicError
from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_cons_strategies import LocalAuthStrategy, AbstractAuthStrategy, ConfigLedgerAuthStrategy
from indy_common.authorize.auth_constraints import AuthConstraint, ConstraintsSerializer
from plenum.common.constants import TRUSTEE, STEWARD
from state.pruning_state import PruningState
from storage.kv_in_memory import KeyValueStorageInMemory


@pytest.fixture(scope='function')
def auth_map(action_add, action_edit):
    """
    action_ids:
    for add: "ADD--SomeType--some_field--*--new_value"
    for edit: "EDIT--SomeType--some_field--old_value--new_value"
    """
    return {action_add.get_action_id(): AuthConstraint(role=STEWARD,
                                                       sig_count=3),
            action_edit.get_action_id(): AuthConstraint(role=STEWARD,
                                                        sig_count=2)}


@pytest.fixture(scope='function')
def local_auth_strategy(auth_map):
    return LocalAuthStrategy(auth_map=auth_map)


def test_local_strategy_found_the_same_action_id(local_auth_strategy):
    assert local_auth_strategy.get_auth_constraint("SomeType--ADD--some_field--*--new_value")
    assert local_auth_strategy.get_auth_constraint("SomeType--EDIT--some_field--old_value--new_value")


def test_local_strategy_not_found_action_id(local_auth_strategy):
    assert local_auth_strategy.get_auth_constraint("SomeType--ADD--some_field--*--other_new_value") is None
    assert local_auth_strategy.get_auth_constraint("SomeType--EDIT--some_field--old_value--other_new_value") is None


"""Group of tests for checking is_accepted method"""


@pytest.fixture(scope='module')
def is_accepted():
    return AbstractAuthStrategy.is_accepted_action_id


@pytest.fixture
def config_ledger_strategy(state, state_serializer):
    return ConfigLedgerAuthStrategy(auth_map,
                                    state=state,
                                    serializer=state_serializer)


def test_is_accepted_by_the_same(is_accepted):
    assert is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                       "TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_not_is_accepted_by_prefix(is_accepted):
    assert not is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                           "aaTYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_not_is_accepted_by_type(is_accepted):
    assert not is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                           "TYPE--aaPREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_not_is_accepted_by_field(is_accepted):
    assert not is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                           "TYPE--PREFIX--aaFIELD--OLD_VALUE--NEW_VALUE")


def test_not_is_accepted_by_old_value(is_accepted):
    assert not is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                           "TYPE--PREFIX--FIELD--aaOLD_VALUE--NEW_VALUE")


def test_not_is_accepted_by_new_value(is_accepted):
    assert not is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE",
                           "TYPE--PREFIX--FIELD--OLD_VALUE--aaNEW_VALUE")


def test_is_accepted_for_all_field(is_accepted):
    assert is_accepted("TYPE--PREFIX--*--OLD_VALUE--NEW_VALUE",
                       "TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_is_accepted_for_all_old_value(is_accepted):
    assert is_accepted("TYPE--PREFIX--FIELD--*--NEW_VALUE",
                       "TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_is_accepted_for_all_new_value(is_accepted):
    assert is_accepted("TYPE--PREFIX--FIELD--OLD_VALUE--*",
                       "TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")


def test_is_accepted_for_all(is_accepted):
    assert is_accepted("TYPE--PREFIX--*--*--*",
                       "TYPE--PREFIX--FIELD--OLD_VALUE--NEW_VALUE")



import pytest

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import ROLE, AuthConstraint
from indy_common.constants import NYM
from indy_node.test.auth_rule.helper import create_verkey_did, sdk_send_and_check_auth_rule_request
from plenum.common.constants import STEWARD, STEWARD_STRING
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def test_check_rule_for_add_action_changing(looper,
                                            sdk_wallet_trustee,
                                            sdk_wallet_steward,
                                            sdk_pool_handle):
    wh, _ = sdk_wallet_trustee
    did1, verkey1 = create_verkey_did(looper, wh)
    """Adding new steward for old auth rules"""
    sdk_add_new_nym(looper,
                    sdk_pool_handle,
                    sdk_wallet_trustee,
                    'newSteward1',
                    STEWARD_STRING,
                    dest=did1, verkey=verkey1)

    constraint = AuthConstraint(role=STEWARD,
                                sig_count=1)
    sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=NYM, field=ROLE, new_value=STEWARD, old_value=None,
                                         constraint=constraint.as_dict)

    did2, verkey2 = create_verkey_did(looper, wh)
    with pytest.raises(RequestRejectedException, match="Not enough STEWARD signatures"):
        """Adding new steward for changed auth rules"""
        sdk_add_new_nym(looper,
                        sdk_pool_handle,
                        sdk_wallet_trustee,
                        'newSteward2',
                        STEWARD_STRING,
                        dest=did2, verkey=verkey2)
    """We change auth rules and for now only steward can add new steward"""
    sdk_add_new_nym(looper,
                    sdk_pool_handle,
                    sdk_wallet_steward,
                    'newSteward2',
                    STEWARD_STRING,
                    dest=did2, verkey=verkey2)



from indy.ledger import append_request_endorser

from plenum.test.helper import sdk_get_and_check_replies, sdk_send_signed_requests, sdk_multisign_request_object


def sdk_append_request_endorser(
        looper,
        request_json,
        endorser_did):
    return looper.loop.run_until_complete(append_request_endorser(request_json, endorser_did))


def sdk_submit_and_check_by_endorser(looper, sdk_pool_handle, sdk_wallet_author, sdk_wallet_endorser, request_json):
    _, endorser_did = sdk_wallet_endorser

    # 1. append Endorser field
    request_json = sdk_append_request_endorser(looper, request_json, endorser_did)

    # 2. sign by both Author and Endorser
    request_json = sdk_multisign_request_object(looper, sdk_wallet_author, request_json)
    request_json = sdk_multisign_request_object(looper, sdk_wallet_endorser, request_json)

    # 3. submit by Endorser
    request_couple = sdk_send_signed_requests(sdk_pool_handle, [request_json])[0]
    sdk_get_and_check_replies(looper, [request_couple])



from typing import Optional

from indy_common.constants import NYM, ROLE
from indy_common.types import Request
from plenum.common.constants import TARGET_NYM, VERKEY, TXN_TIME, DOMAIN_LEDGER_ID
from plenum.common.txn_util import get_txn_time, get_payload_data, get_seq_no, append_txn_metadata
from plenum.common.types import f
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler


class IdrCacheNymHandler(WriteRequestHandler):
    def __init__(self, database_manager: DatabaseManager):
        super().__init__(database_manager, NYM, DOMAIN_LEDGER_ID)

    def apply_request(self, request: Request, batch_ts, prev_result):
        txn = self._req_to_txn(request)
        txn = append_txn_metadata(txn,
                                  seq_no=prev_result.get(f.SEQ_NO.nm),
                                  txn_time=prev_result.get(TXN_TIME))
        return self.update_state(txn, prev_result, request)

    def update_state(self, txn, prev_result, request, is_committed=False):
        txn_time = get_txn_time(txn)
        nym = get_payload_data(txn).get(TARGET_NYM)
        seq_no = get_seq_no(txn)
        self.database_manager.idr_cache.set(nym,
                                            seqNo=seq_no,
                                            txnTime=txn_time,
                                            ta=prev_result.get(f.IDENTIFIER.nm),
                                            role=prev_result.get(ROLE),
                                            verkey=prev_result.get(VERKEY),
                                            isCommitted=is_committed)
        return seq_no, txn_time, txn

    def static_validation(self, request):
        pass

    def additional_dynamic_validation(self, request, req_pp_time: Optional[int]):
        pass

    def gen_state_key(self, txn):
        pass



import json
import pytest

from indy.did import create_and_store_my_did, replace_keys_start, replace_keys_apply
from indy.ledger import build_get_nym_request
from indy_common.constants import ENDORSER_STRING
from indy_node.test.helper import sdk_add_attribute_and_check
from plenum.common.exceptions import RequestRejectedException
from plenum.common.util import randomString
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request, \
    prepare_nym_request, sdk_add_new_nym

ENDORSER_SEED = 'TRUST0NO0ONE00000000000000000001'


def set_verkey(looper, sdk_pool_handle, sdk_wallet_sender, dest, verkey):
    wh, _ = sdk_wallet_sender
    nym_request, new_did = looper.loop.run_until_complete(
        prepare_nym_request(sdk_wallet_sender, None,
                            None, ENDORSER_STRING, dest, verkey, False if verkey else True))
    request_couple = sdk_sign_and_send_prepared_request(looper, sdk_wallet_sender,
                                                        sdk_pool_handle, nym_request)
    sdk_get_and_check_replies(looper, [request_couple])
    return wh, new_did


@pytest.fixture(scope="module")
def endorser_did_verkey(looper, sdk_wallet_client):
    wh, _ = sdk_wallet_client
    named_did, verkey = looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': ENDORSER_SEED})))
    return named_did, verkey


def test_pool_nodes_started(nodeSet):
    pass

@pytest.fixture(scope='function', params=['trustee', 'steward'])
def sdk_wallet(request, sdk_wallet_steward, sdk_wallet_trustee):
    if request.param == 'steward':
        yield sdk_wallet_steward
    elif request.param == 'trustee':
        yield sdk_wallet_trustee


def test_send_same_nyms_only_first_gets_written(
        looper, sdk_pool_handle, sdk_wallet):
    wh, _ = sdk_wallet
    seed = randomString(32)
    did, verkey = looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': seed})))

    # request 1
    _, did1 = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet, dest=did, verkey=verkey)

    seed = randomString(32)
    _, verkey = looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': seed})))
    # request 2
    with pytest.raises(RequestRejectedException) as e:
        _, did2 = sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet, dest=did, verkey=verkey)
    e.match('can not touch verkey field since only the owner can modify it')


def get_nym(looper, sdk_pool_handle, sdk_wallet_steward, t_did):
    _, s_did = sdk_wallet_steward
    get_nym_req = looper.loop.run_until_complete(build_get_nym_request(s_did, t_did))
    req = sdk_sign_and_send_prepared_request(looper, sdk_wallet_steward,
                                             sdk_pool_handle, get_nym_req)
    return sdk_get_and_check_replies(looper, [req])


def test_get_nym_without_adding_it(looper, sdk_pool_handle, sdk_wallet_steward,
                                   endorser_did_verkey):
    t_did, _ = endorser_did_verkey
    rep = get_nym(looper, sdk_pool_handle, sdk_wallet_steward, t_did)
    assert not rep[0][1]['result']['data']


@pytest.fixture(scope="module")
def nym_added(looper, sdk_pool_handle, sdk_wallet_steward, endorser_did_verkey):
    dest, _ = endorser_did_verkey
    set_verkey(looper, sdk_pool_handle, sdk_wallet_steward, dest, None)


def test_add_nym(nym_added):
    pass


def test_get_nym_without_verkey(looper, sdk_pool_handle, sdk_wallet_steward, nym_added,
                                endorser_did_verkey):
    t_did, _ = endorser_did_verkey
    rep = get_nym(looper, sdk_pool_handle, sdk_wallet_steward, t_did)
    assert rep[0][1]['result']['data']
    assert not json.loads(rep[0][1]['result']['data'])['verkey']


@pytest.fixture(scope="module")
def verkey_added_to_nym(looper, sdk_pool_handle, sdk_wallet_steward, nym_added, endorser_did_verkey):
    wh, _ = sdk_wallet_steward
    did, _ = endorser_did_verkey
    verkey = looper.loop.run_until_complete(replace_keys_start(wh, did, json.dumps({'': ''})))
    set_verkey(looper, sdk_pool_handle, sdk_wallet_steward, did, verkey)
    looper.loop.run_until_complete(replace_keys_apply(wh, did))


def test_add_verkey_to_existing_nym(verkey_added_to_nym):
    pass


def test_get_did_with_verkey(looper, sdk_pool_handle, sdk_wallet_steward, verkey_added_to_nym,
                             endorser_did_verkey):
    t_did, _ = endorser_did_verkey
    rep = get_nym(looper, sdk_pool_handle, sdk_wallet_steward, t_did)
    assert rep[0][1]['result']['data']
    assert json.loads(rep[0][1]['result']['data'])['verkey']


def test_send_attrib_for_did(looper, sdk_pool_handle, sdk_wallet_steward,
                             verkey_added_to_nym, endorser_did_verkey):
    raw = '{"name": "Alice"}'
    dest, _ = endorser_did_verkey
    wh, _ = sdk_wallet_steward
    sdk_add_attribute_and_check(looper, sdk_pool_handle, (wh, dest), raw, dest)


@pytest.fixture(scope="module")
def verkey_removed_from_existing_did(looper, sdk_pool_handle, sdk_wallet_steward,
                                     verkey_added_to_nym, endorser_did_verkey):
    did, _ = endorser_did_verkey
    wh, _ = sdk_wallet_steward
    set_verkey(looper, sdk_pool_handle, (wh, did), did, None)


def test_remove_verkey_from_did(verkey_removed_from_existing_did):
    pass


@pytest.mark.skip(
    reason="SOV-568. Obsolete assumption, if an identity has set "
           "its verkey to blank, no-one including "
           "itself can change it")
def testNewverkey_added_to_nym(be, do, philCli, abbrevIdr,
                               verkeyRemovedFromExistingDID):
    pass



import pytest
from plenum.common.constants import STEWARD

from indy_common.constants import CONSTRAINT, AUTH_ACTION
from indy_common.types import ClientAuthRuleOperation
from indy_node.test.helper import generate_constraint_entity
from indy_node.test.auth_rule.helper import generate_constraint_list, \
    generate_auth_rule_operation
from indy_common.authorize.auth_constraints import ROLE, AUTH_CONSTRAINTS, CONSTRAINT_ID, AuthConstraint, SIG_COUNT, \
    NEED_TO_BE_OWNER, OFF_LEDGER_SIGNATURE, METADATA, ConstraintsEnum

validator = ClientAuthRuleOperation()

valid_auth_rule_operation_small = generate_auth_rule_operation()
valid_auth_rule_operation_large = generate_auth_rule_operation(constraint=generate_constraint_list(
    auth_constraints=[generate_constraint_entity(),
                      generate_constraint_entity()]))
valid_auth_rule_operation_extra_large = generate_auth_rule_operation(constraint=generate_constraint_list(
    auth_constraints=[generate_constraint_entity(),
                      generate_constraint_list(
                          auth_constraints=[generate_constraint_entity(),
                                            generate_constraint_entity()])]))


def test_valid():
    validator.validate(valid_auth_rule_operation_small)
    validator.validate(valid_auth_rule_operation_large)
    validator.validate(valid_auth_rule_operation_extra_large)


def test_invalid_operation_action():
    # must be ADD_PREFIX or EDIT_PREFIX
    invalid_auth_rule_operation = generate_auth_rule_operation(auth_action="auth_action")
    with pytest.raises(TypeError, match=AUTH_ACTION):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_entity_role():
    # ConstraintEntityField without required field 'role'
    invalid_auth_rule_operation = generate_auth_rule_operation()
    del invalid_auth_rule_operation[CONSTRAINT][ROLE]
    with pytest.raises(TypeError, match=ROLE):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_entity_role_in_extra_large_constraint():
    # ConstraintEntityField without required field 'role'
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=generate_constraint_list(
        auth_constraints=[generate_constraint_entity(),
                          generate_constraint_list(
                              auth_constraints=[generate_constraint_entity(),
                                                generate_constraint_entity()])]))
    del invalid_auth_rule_operation[CONSTRAINT][AUTH_CONSTRAINTS][1][AUTH_CONSTRAINTS][0][ROLE]
    with pytest.raises(TypeError, match=ROLE):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_auth_constraints():
    # ConstraintListField without required field 'auth_constraints'
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=generate_constraint_list(
        auth_constraints=[generate_constraint_entity(),
                          generate_constraint_entity()]))
    del invalid_auth_rule_operation[CONSTRAINT][AUTH_CONSTRAINTS]
    with pytest.raises(TypeError, match=AUTH_CONSTRAINTS):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_auth_constraints_with_large_constraint():
    # ConstraintListField without required field 'auth_constraints' on the 2nd level
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=generate_constraint_list(
        auth_constraints=[generate_constraint_entity(),
                          generate_constraint_list(
                              auth_constraints=[generate_constraint_entity(),
                                                generate_constraint_entity()])]))
    del invalid_auth_rule_operation[CONSTRAINT][AUTH_CONSTRAINTS][1][AUTH_CONSTRAINTS]
    with pytest.raises(TypeError, match=AUTH_CONSTRAINTS):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_constraint_id_with_large_constraint():
    # ConstraintListField without required field CONSTRAINT_ID on the 2nd level
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=generate_constraint_list(
        auth_constraints=[generate_constraint_entity(),
                          generate_constraint_list(
                              auth_constraints=[generate_constraint_entity(),
                                                generate_constraint_entity()])]))
    del invalid_auth_rule_operation[CONSTRAINT][AUTH_CONSTRAINTS][1][CONSTRAINT_ID]
    with pytest.raises(TypeError, match=CONSTRAINT_ID):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_with_empty_auth_constraints():
    # ConstraintListField without empty list in auth_constraints
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=generate_constraint_list(
        auth_constraints=[generate_constraint_entity(),
                          generate_constraint_list(
                              auth_constraints=[])]))
    with pytest.raises(TypeError, match="Fields {} should not be an empty "
                                        "list.".format(AUTH_CONSTRAINTS)):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_with_empty_constraint_list():
    # ClientAuthRuleOperation with empty list of constraints
    invalid_auth_rule_operation = generate_auth_rule_operation(constraint=[])
    with pytest.raises(TypeError, match="Fields {} and {} are required and should not "
                                        "be an empty list.".format(AUTH_CONSTRAINTS, CONSTRAINT)):
        validator.validate(invalid_auth_rule_operation)


def test_invalid_operation_without_none_constraint():
    # ConstraintListField without None in field CONSTRAINT
    invalid_auth_rule_operation = generate_auth_rule_operation()
    invalid_auth_rule_operation[CONSTRAINT] = None
    with pytest.raises(TypeError, match="Fields {} and {} are required and should not "
                                        "be an empty list.".format(AUTH_CONSTRAINTS, CONSTRAINT)):
        validator.validate(invalid_auth_rule_operation)


def test_auth_constraint_from_dct_succesfull():
    unkwn = 'unknown_field'
    auth_dct = {ROLE: STEWARD,
                SIG_COUNT: 1,
                NEED_TO_BE_OWNER: False,
                OFF_LEDGER_SIGNATURE: False,
                METADATA: {},
                unkwn: 'unknown_value'}
    constr = AuthConstraint.from_dict(auth_dct)

    for k in auth_dct:
        if k != unkwn:
            assert hasattr(constr, k)
    assert not hasattr(constr, unkwn)


def test_auth_constraint_without_off_ledger_sig_from_dct_succesfull():
    unkwn = 'unknown_field'
    auth_dct = {CONSTRAINT_ID: ConstraintsEnum.ROLE_CONSTRAINT_ID,
                ROLE: STEWARD,
                SIG_COUNT: 1,
                NEED_TO_BE_OWNER: False,
                METADATA: {},
                unkwn: 'unknown_value'}
    constr = AuthConstraint.from_dict(auth_dct)

    for k in auth_dct:
        if k != unkwn:
            assert hasattr(constr, k)
    assert not hasattr(constr, unkwn)
    assert constr.off_ledger_signature is None

    del auth_dct[unkwn]
    assert auth_dct == constr.as_dict



import base64
import copy
import random
import time

import base58

from common.serializers.serialization import state_roots_serializer, domain_state_serializer
from crypto.bls.bls_multi_signature import MultiSignature, MultiSignatureValue
from indy_common.constants import \
    ATTRIB, CLAIM_DEF, SCHEMA, CLAIM_DEF_FROM, CLAIM_DEF_SCHEMA_REF, CLAIM_DEF_SIGNATURE_TYPE, \
    CLAIM_DEF_PUBLIC_KEYS, CLAIM_DEF_TAG, SCHEMA_NAME, SCHEMA_VERSION, SCHEMA_ATTR_NAMES, GET_NYM, GET_ATTR, \
    GET_CLAIM_DEF, GET_SCHEMA
from indy_common.state import domain
from indy_common.types import Request
from plenum.common.constants import TXN_TYPE, TARGET_NYM, RAW, DATA, \
    IDENTIFIER, NAME, VERSION, ROLE, VERKEY, STATE_PROOF, ROOT_HASH, MULTI_SIGNATURE, PROOF_NODES, TXN_TIME, \
    CURRENT_PROTOCOL_VERSION, DOMAIN_LEDGER_ID, NYM
from plenum.common.txn_util import reqToTxn, append_txn_metadata, append_payload_metadata, set_type
from plenum.common.types import f
from plenum.common.util import get_utc_epoch, friendlyToHex, hexToFriendly
from plenum.server.request_handlers.utils import nym_to_state_key


def extract_proof(result, expected_multi_sig):
    proof = result[STATE_PROOF]
    assert proof
    assert proof[ROOT_HASH]
    assert proof[PROOF_NODES]
    multi_sign = proof[MULTI_SIGNATURE]
    assert multi_sign
    assert multi_sign == expected_multi_sig
    return proof


def save_multi_sig(db_manager):
    multi_sig_value = MultiSignatureValue(ledger_id=DOMAIN_LEDGER_ID,
                                          state_root_hash=state_roots_serializer.serialize(
                                              bytes(db_manager.get_state(DOMAIN_LEDGER_ID).committedHeadHash)),
                                          txn_root_hash='2' * 32,
                                          pool_state_root_hash='1' * 32,
                                          timestamp=get_utc_epoch())
    multi_sig = MultiSignature('0' * 32, ['Alpha', 'Beta', 'Gamma'], multi_sig_value)
    db_manager.bls_store.put(multi_sig)
    return multi_sig.as_dict()


def is_proof_verified(db_manager,
                      proof, path,
                      value, seq_no, txn_time):
    encoded_value = domain.encode_state_value(value, seq_no, txn_time)
    proof_nodes = base64.b64decode(proof[PROOF_NODES])
    root_hash = base58.b58decode(proof[ROOT_HASH])
    verified = db_manager.get_state(DOMAIN_LEDGER_ID).verify_state_proof(
        root_hash,
        path,
        encoded_value,
        proof_nodes,
        serialized=True
    )
    return verified


# Similar tests for Rich Schema objects are in indy_node/test/request_handlers/rich_schema

def test_state_proofs_for_get_attr(write_manager,
                                   read_manager,
                                   db_manager):
    # Adding attribute
    nym = 'Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv'
    attr_key = 'last_name'
    raw_attribute = '{"last_name":"Anderson"}'
    seq_no = 0
    txn_time = int(time.time())
    identifier = "6ouriXMZkLeHsuXrN1X1fd"
    txn = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: nym,
        RAW: raw_attribute,
    }
    txn = append_txn_metadata(reqToTxn(Request(operation=txn,
                                               protocolVersion=CURRENT_PROTOCOL_VERSION,
                                               identifier=identifier)),
                              seq_no=seq_no, txn_time=txn_time)
    write_manager.update_state(txn)
    db_manager.get_state(DOMAIN_LEDGER_ID).commit()
    multi_sig = save_multi_sig(db_manager)

    # Getting attribute
    get_request = Request(
        operation={
            TARGET_NYM: nym,
            RAW: 'last_name',
            TXN_TYPE: GET_ATTR,
        },
        signatures={},
        protocolVersion=CURRENT_PROTOCOL_VERSION
    )
    result = read_manager.get_result(get_request)

    proof = extract_proof(result, multi_sig)
    attr_value = result[DATA]
    assert attr_value == raw_attribute

    # Verifying signed state proof
    path = domain.make_state_path_for_attr(nym, attr_key)
    assert is_proof_verified(db_manager,
                             proof, path,
                             domain.hash_of(attr_value), seq_no, txn_time)


def test_state_proofs_for_get_claim_def(write_manager,
                                        read_manager,
                                        db_manager):
    # Adding claim def
    nym = 'Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv'

    seq_no = 0
    txn_time = int(time.time())
    identifier = "6ouriXMZkLeHsuXrN1X1fd"

    schema_seqno = 0
    signature_type = 'CL'
    key_components = '{"key_components": []}'
    tag = 'tag1'

    txn = {
        TXN_TYPE: CLAIM_DEF,
        TARGET_NYM: nym,
        CLAIM_DEF_SCHEMA_REF: schema_seqno,
        CLAIM_DEF_PUBLIC_KEYS: key_components,
        CLAIM_DEF_TAG: tag
    }
    txn = append_txn_metadata(reqToTxn(Request(operation=txn,
                                               protocolVersion=CURRENT_PROTOCOL_VERSION,
                                               identifier=identifier)),
                              seq_no=seq_no, txn_time=txn_time)
    txn = append_payload_metadata(txn, frm=nym)

    write_manager.update_state(txn)
    db_manager.get_state(DOMAIN_LEDGER_ID).commit()
    multi_sig = save_multi_sig(db_manager)

    # Getting claim def
    request = Request(
        operation={
            IDENTIFIER: nym,
            CLAIM_DEF_FROM: nym,
            CLAIM_DEF_SCHEMA_REF: schema_seqno,
            CLAIM_DEF_SIGNATURE_TYPE: signature_type,
            CLAIM_DEF_TAG: tag,
            TXN_TYPE: GET_CLAIM_DEF,
        },
        signatures={},
        protocolVersion=CURRENT_PROTOCOL_VERSION
    )

    result = read_manager.get_result(request)
    proof = extract_proof(result, multi_sig)
    assert result[DATA] == key_components

    # Verifying signed state proof
    path = domain.make_state_path_for_claim_def(nym, schema_seqno,
                                                signature_type, tag)
    assert is_proof_verified(db_manager,
                             proof, path,
                             key_components, seq_no, txn_time)


def test_state_proofs_for_get_schema(write_manager,
                                     read_manager,
                                     db_manager):
    # Adding schema
    nym = 'Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv'

    seq_no = 0
    txn_time = int(time.time())
    identifier = "6ouriXMZkLeHsuXrN1X1fd"

    schema_name = "schema_a"
    schema_version = "1.0"
    # data = '{"name": "schema_a", "version": "1.0"}'
    schema_key = {SCHEMA_NAME: schema_name,
                  SCHEMA_VERSION: schema_version}
    data = {**schema_key,
            SCHEMA_ATTR_NAMES: ["Some_Attr", "Attr1"]}
    txn = {
        TXN_TYPE: SCHEMA,
        DATA: data,
    }
    txn = append_txn_metadata(reqToTxn(Request(operation=txn,
                                               protocolVersion=CURRENT_PROTOCOL_VERSION,
                                               identifier=identifier)),
                              seq_no=seq_no, txn_time=txn_time)
    txn = append_payload_metadata(txn, frm=nym)

    write_manager.update_state(txn)
    db_manager.get_state(DOMAIN_LEDGER_ID).commit()
    multi_sig = save_multi_sig(db_manager)

    # Getting schema
    request = Request(
        operation={
            TARGET_NYM: nym,
            DATA: schema_key,
            TXN_TYPE: GET_SCHEMA
        },
        signatures={},
        protocolVersion=CURRENT_PROTOCOL_VERSION
    )

    result = read_manager.get_result(request)
    proof = extract_proof(result, multi_sig)
    assert result[DATA] == data

    data.pop(NAME)
    data.pop(VERSION)

    # Verifying signed state proof
    path = domain.make_state_path_for_schema(nym, schema_name, schema_version)
    assert is_proof_verified(db_manager,
                             proof, path,
                             data, seq_no, txn_time)


def prep_multi_sig(write_manager, db_manager, nym, role, verkey, seq_no):
    txn_time = int(time.time())
    identifier = "6ouriXMZkLeHsuXrN1X1fd"

    # Adding nym
    data = {
        TARGET_NYM: nym,
        ROLE: role,
        VERKEY: verkey,
    }
    txn = append_txn_metadata(reqToTxn(Request(operation=data,
                                               protocolVersion=CURRENT_PROTOCOL_VERSION,
                                               identifier=identifier)),
                              seq_no=seq_no, txn_time=txn_time)
    txn = set_type(txn, NYM)
    txn = append_payload_metadata(txn, frm=nym)
    write_manager.update_state(txn)
    db_manager.get_state(DOMAIN_LEDGER_ID).commit()
    multi_sig = save_multi_sig(db_manager)
    return data, multi_sig


def get_nym_verify_proof(read_manager,
                         db_manager,
                         nym, data, multi_sig):
    request = Request(
        operation={
            TARGET_NYM: nym,
            TXN_TYPE: GET_NYM,
        },
        signatures={},
        protocolVersion=CURRENT_PROTOCOL_VERSION
    )
    result = read_manager.get_result(request)
    proof = extract_proof(result, multi_sig)
    result_data = None

    assert proof
    if data:
        assert result[DATA]
        result_data = domain_state_serializer.deserialize(result[DATA])
        data_from_req = copy.copy(result_data)
        data_from_req.pop(f.IDENTIFIER.nm, None)
        data_from_req.pop(f.SEQ_NO.nm, None)
        data_from_req.pop(TXN_TIME, None)
        assert data_from_req == data

    # Verifying signed state proof
    path = nym_to_state_key(nym)
    # If the value does not exist, serialization should be null and
    # verify_state_proof needs to be given null (None). This is done to
    # differentiate between absence of value and presence of empty string value
    if result_data:
        data.pop(TARGET_NYM, None)
        data.update({f.IDENTIFIER.nm: result_data[f.IDENTIFIER.nm],
                     f.SEQ_NO.nm: result_data[f.SEQ_NO.nm],
                     TXN_TIME: result_data[TXN_TIME]})
    serialized_value = domain_state_serializer.serialize(data) if data else None
    proof_nodes = base64.b64decode(proof[PROOF_NODES])
    root_hash = base58.b58decode(proof[ROOT_HASH])
    return db_manager.get_state(DOMAIN_LEDGER_ID).verify_state_proof(
        root_hash,
        path,
        serialized_value,
        proof_nodes,
        serialized=True
    )


def test_state_proofs_for_get_nym(write_manager,
                                  read_manager,
                                  db_manager):
    nym = 'Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv'
    role = "2"
    verkey = "~7TYfekw4GUagBnBVCqPjiC"
    seq_no = 1
    # Check for existing nym
    data, multi_sig = prep_multi_sig(write_manager, db_manager, nym, role, verkey, seq_no)
    assert get_nym_verify_proof(read_manager, db_manager, nym, data, multi_sig)

    # Shuffle the bytes of nym
    h = list(friendlyToHex(nym))
    random.shuffle(h)
    garbled_nym = hexToFriendly(bytes(h))
    data[f.IDENTIFIER.nm] = garbled_nym
    # `garbled_nym` does not exist, proof should verify but data is null
    assert get_nym_verify_proof(read_manager, db_manager, garbled_nym, None, multi_sig)


def test_no_state_proofs_if_protocol_version_less(write_manager,
                                                  read_manager,
                                                  db_manager):
    nym = 'Gw6pDLhcBcoQesN72qfotTgFa7cbuqZpkX3Xo6pLhPhv'
    role = "2"
    verkey = "~7TYfekw4GUagBnBVCqPjiC"
    identifier = "6ouriXMZkLeHsuXrN1X1fd"

    seq_no = 0
    txn_time = int(time.time())
    # Adding nym
    data = {
        TARGET_NYM: nym,
        ROLE: role,
        VERKEY: verkey,
    }
    txn = append_txn_metadata(reqToTxn(Request(operation=data,
                                               protocolVersion=CURRENT_PROTOCOL_VERSION,
                                               identifier=identifier)),
                              seq_no=seq_no, txn_time=txn_time)
    txn = set_type(txn, NYM)
    txn = append_payload_metadata(txn, frm=nym)
    write_manager.update_state(txn)
    db_manager.get_state(DOMAIN_LEDGER_ID).commit()
    multi_sig = save_multi_sig(db_manager)

    # Getting nym
    request = Request(
        operation={
            TARGET_NYM: nym,
            TXN_TYPE: GET_NYM,
        },
        signatures={}
    )
    result = read_manager.get_result(request)
    assert STATE_PROOF not in result



from plenum.common.constants import TXN_AUTHOR_AGREEMENT, TXN_AUTHOR_AGREEMENT_DIGEST, \
    TXN_AUTHOR_AGREEMENT_RATIFICATION_TS, TXN_AUTHOR_AGREEMENT_VERSION, TXN_AUTHOR_AGREEMENT_RETIREMENT_TS, \
    CURRENT_TXN_PAYLOAD_VERSIONS
import time

from indy_node.test.helper import start_stopped_node
from plenum.common.txn_util import get_txn_time
from plenum.common.util import randomString, get_utc_epoch
from plenum.test.node_catchup.helper import ensure_all_nodes_have_same_data
from plenum.test.pool_transactions.helper import disconnect_node_and_ensure_disconnected
from plenum.test.txn_author_agreement.helper import sdk_send_txn_author_agreement, sdk_get_txn_author_agreement, \
    sdk_send_txn_author_agreement_disable


def test_recover_taa_from_ledger(txnPoolNodeSet,
                                 sdk_pool_handle,
                                 sdk_wallet_trustee,
                                 looper,
                                 monkeypatch,
                                 setup_aml,
                                 tconf,
                                 tdir,
                                 allPluginsPath):
    orig_handlers = {}
    # Step 1. Stop one node
    node_to_stop = txnPoolNodeSet[-1]
    rest_pool = txnPoolNodeSet[:-1]
    disconnect_node_and_ensure_disconnected(looper,
                                            txnPoolNodeSet,
                                            node_to_stop.name,
                                            stopNode=True)
    looper.removeProdable(name=node_to_stop.name)

    # Step 2. Patch all the rest nodes for using old version TAA handler

    # it's ugly but it works
    globals()['CURRENT_TXN_PAYLOAD_VERSIONS'][TXN_AUTHOR_AGREEMENT] = '1'
    for node in rest_pool:
        handler = node.write_manager.request_handlers.get(TXN_AUTHOR_AGREEMENT)[0]
        orig_handlers[node.name] = handler
        handler_for_v_1 = node.write_manager._request_handlers_with_version.get((TXN_AUTHOR_AGREEMENT, "1"))[0]
        node.write_manager.request_handlers[TXN_AUTHOR_AGREEMENT] = [handler_for_v_1]

    # Step 3. Send TAA txn in old way
    text = randomString(1024)
    version_0 = randomString(16)
    res = sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_0, text)[1]
    taa_0_ratification_ts = get_txn_time(res['result'])

    version_1 = randomString(16)
    res = sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_1, text)[1]
    taa_1_ratification_ts = get_txn_time(res['result'])

    # Step 4. return original TAA handlers back

    # it's ugly but it works
    globals()['CURRENT_TXN_PAYLOAD_VERSIONS'][TXN_AUTHOR_AGREEMENT] = '2'
    for node in rest_pool:
        node.write_manager.request_handlers[TXN_AUTHOR_AGREEMENT] = [orig_handlers[node.name]]

    # Step 5. Send another TAA txn in new way without optional parameters
    text_2 = randomString(1024)
    version_2 = randomString(16)
    ratified_2 = get_utc_epoch() - 300
    res_0 = sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee,
                                          version_2, text_2, ratified=ratified_2)[1]

    # Step 6. Send another TAA txn in new way without optional parameter
    text = randomString(1024)
    version_3 = randomString(16)
    ratified_3 = get_utc_epoch() - 300
    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_3, text, ratified=ratified_3)

    # Step 7. Send taa updating for the second taa transaction (for checking txn with optional parameter)
    retired_time = int(time.time()) + 20
    retired_time_in_past = int(time.time()) - 20
    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_2, retired=retired_time)

    # Step 8. Ensure, that all TAAs was written
    res_1 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_1)[1]
    assert TXN_AUTHOR_AGREEMENT_DIGEST not in res_1['result']['data']
    assert TXN_AUTHOR_AGREEMENT_RATIFICATION_TS not in res_1['result']['data']
    assert res_1['result']['data'][TXN_AUTHOR_AGREEMENT_VERSION] == version_1

    res_2 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_2)[1]
    check_result_contains_expected_taa_data(res_2, version_2, ratified_2, retired_time)

    res_3 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_3)[1]
    check_result_contains_expected_taa_data(res_3, version_3, ratified_3)

    # Step 9. Return previous disconnected node back
    node_to_stop = start_stopped_node(node_to_stop, looper,
                                      tconf, tdir, allPluginsPath)
    txnPoolNodeSet = rest_pool + [node_to_stop]

    # Step 10. Ensure that all nodes have the same data
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)

    # Step 11. Send another taa txns for checking pool writability
    text = randomString(1024)
    version_4 = randomString(16)
    ratified_4 = get_utc_epoch() - 300
    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_4, text, ratified=ratified_4)

    # Step 12. Ensure that all nodes have the same data
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)

    # Step 13. Retire TAA written using old handler, make sure ratification date is not spoiled
    sdk_send_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version_1, retired=retired_time_in_past)
    res_1 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_1)[1]
    check_result_contains_expected_taa_data(res_1, version_1, taa_1_ratification_ts, retired_time_in_past)

    # Step 14. Disable TAAs written using old handler, make sure ratification date is not spoiled
    disable_res = sdk_send_txn_author_agreement_disable(looper, sdk_pool_handle, sdk_wallet_trustee)[1]

    res_0 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_0)[1]
    check_result_contains_expected_taa_data(res_0, version_0, taa_0_ratification_ts, get_txn_time(disable_res['result']))

    res_1 = sdk_get_txn_author_agreement(looper, sdk_pool_handle, sdk_wallet_trustee, version=version_1)[1]
    check_result_contains_expected_taa_data(res_1, version_1, taa_1_ratification_ts, retired_time_in_past)


def check_result_contains_expected_taa_data(result, version, ratification_ts, retired_time=None):
    assert TXN_AUTHOR_AGREEMENT_DIGEST in result['result']['data']
    assert TXN_AUTHOR_AGREEMENT_RATIFICATION_TS in result['result']['data']
    assert result['result']['data'][TXN_AUTHOR_AGREEMENT_VERSION] == version
    assert result['result']['data'][TXN_AUTHOR_AGREEMENT_RATIFICATION_TS] == ratification_ts
    if retired_time:
        assert result['result']['data'][TXN_AUTHOR_AGREEMENT_RETIREMENT_TS] == retired_time
    else:
        assert TXN_AUTHOR_AGREEMENT_RETIREMENT_TS not in result['result']['data']




#!/usr/bin/env python

import os
import sys

from setuptools import setup, find_packages

v = sys.version_info
if sys.version_info < (3, 8):
    msg = "FAIL: Requires Python 3.8 or later, " \
          "but setup.py was run using {}.{}.{}"
    v = sys.version_info
    print(msg.format(v.major, v.minor, v.micro))
    # noinspection PyPackageRequirements
    print("NOTE: Installation failed. Run setup.py using python3")
    sys.exit(1)

try:
    here = os.path.abspath(os.path.dirname(__file__))
except NameError:
    # it can be the case when we are being run as script or frozen
    here = os.path.abspath(os.path.dirname(sys.argv[0]))

metadata = {'__file__': os.path.join(here, 'indy_node', '__metadata__.py')}
with open(metadata['__file__'], 'r') as f:
    exec(f.read(), metadata)

BASE_DIR = os.path.join(os.path.expanduser("~"), ".indy")

tests_require = ['attrs>=20.3.0', 'pytest>=6.2.2', 'pytest-xdist>=2.2.1', 'pytest-forked>=1.3.0',
                 'python3-indy==1.16.0.post236', 'pytest-asyncio>=0.14.0']

setup(
    name=metadata['__title__'],
    version=metadata['__version__'],
    description=metadata['__description__'],
    long_description=metadata['__long_description__'],
    keywords=metadata['__keywords__'],
    url=metadata['__url__'],
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    maintainer=metadata['__maintainer__'],
    license=metadata['__license__'],
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    packages=find_packages(exclude=['docs', 'docs*']) + [
        'data'],
    # TODO move that to MANIFEST.in
    package_data={
        '': ['*.txt', '*.md', '*.rst', '*.json', '*.conf', '*.html',
             '*.css', '*.ico', '*.png', 'LICENSE', 'LEGAL', '*.indy']},
    include_package_data=True,
    data_files=[(
        (BASE_DIR, ['data/nssm_original.exe'])
    )],

    install_requires=['indy-plenum==1.13.0.dev175',
                      'importlib-metadata<3.0',
                      'timeout-decorator>=0.5.0',
                      'distro>=1.5.0'],
    setup_requires=['pytest-runner'],
    extras_require={
        'tests': tests_require
    },
    tests_require=tests_require,
    scripts=['scripts/start_indy_node',
             'scripts/start_node_control_tool',
             'scripts/clear_node.py',
             'scripts/get_keys',
             'scripts/get_metrics',
             'scripts/generate_indy_pool_transactions',
             'scripts/init_indy_keys',
             'scripts/upgrade_indy_node_ubuntu1604.sh',
             'scripts/upgrade_indy_node_ubuntu1604_test.sh',
             'scripts/upgrade_indy_node.bat',
             'scripts/upgrade_indy_node_test.bat',
             'scripts/restart_indy_node_ubuntu1604.sh',
             'scripts/restart_indy_node.bat',
             'scripts/restart_sovrin_node_ubuntu1604.sh',
             'scripts/complete_rebranding_upgrade_ubuntu1604.sh',
             'scripts/install_indy_node.bat',
             'scripts/delete_indy_node.bat',
             'scripts/restart_upgrade_agent.bat',
             'scripts/install_nssm.bat',
             'scripts/read_ledger',
             'scripts/validator-info',
             'scripts/validator-info-history',
             'scripts/init_bls_keys',
             'scripts/create_dirs.sh',
             'scripts/setup_iptables',
             'scripts/setup_indy_node_iptables',
             'scripts/current_validators',
             'scripts/node_address_list',
             'scripts/generate_bls_proof_of_possession',
             'tools/diagnostics/nscapture',
             'tools/diagnostics/nsdiff',
             'tools/diagnostics/nsreplay',
             ]
)



import pytest
from indy_common.types import ClientAttribOperation
from collections import OrderedDict
from plenum.common.messages.fields import ConstantField, LimitedLengthStringField, IdentifierField, \
    JsonField, Sha256HexField


EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("dest", IdentifierField),
    ("raw", JsonField),
    ('enc', LimitedLengthStringField),
    ('hash', Sha256HexField),
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientAttribOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientAttribOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)



import copy
import json

from indy_common.constants import REVOC_REG_DEF_ID, TO, CRED_DEF_ID, REVOC_TYPE, TAG, VALUE, STATE_PROOF_FROM, \
    ACCUM_FROM
from indy_common.state import domain
from indy_node.test.anon_creds.helper import build_get_revoc_reg_delta
from plenum.common.constants import DATA, STATE_PROOF
from plenum.common.types import f, OPERATION
from plenum.common.util import get_utc_epoch
from plenum.test.helper import sdk_send_and_check


def test_send_reg_def_and_get_delta_then(
        looper,
        txnPoolNodeSet,
        sdk_pool_handle,
        send_revoc_reg_def_by_default,
        sdk_wallet_steward):
    rev_def_req, _ = send_revoc_reg_def_by_default
    get_revoc_reg_delta = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    get_revoc_reg_delta['operation'][REVOC_REG_DEF_ID] = domain.make_state_path_for_revoc_def(authors_did=rev_def_req[f.IDENTIFIER.nm],
                                                                                              cred_def_id=rev_def_req[OPERATION][CRED_DEF_ID],
                                                                                              revoc_def_type=rev_def_req[OPERATION][REVOC_TYPE],
                                                                                              revoc_def_tag=rev_def_req[OPERATION][TAG]).decode()
    get_revoc_reg_delta['operation'][TO] = get_utc_epoch()
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg_delta)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert DATA in reply['result']
    assert reply['result'][DATA][STATE_PROOF_FROM] is None
    assert reply['result'][DATA][VALUE][ACCUM_FROM] is None
    assert STATE_PROOF in reply['result']
    assert reply['result'][STATE_PROOF] is not None



import pytest

from indy_common.constants import VALIDATOR_INFO
from plenum.common.constants import TXN_TYPE, DATA
from plenum.common.types import f
from plenum.test.helper import sdk_gen_request, sdk_sign_request_objects


def test_validator_info_handler(monkeypatch,
                                sdk_wallet_trustee,
                                txnPoolNodeSet,
                                looper):

    op = {
        TXN_TYPE: VALIDATOR_INFO
    }
    req_obj = sdk_gen_request(op, identifier=sdk_wallet_trustee[1])
    req_obj.signature = "signature"

    def is_ack(req_key, frm):
        assert (req_obj.identifier, req_obj.reqId) == req_key

    def is_reply_correct(resp, frm):
        _comparison_reply(resp, req_obj)

    node = txnPoolNodeSet[0]
    monkeypatch.setattr(node, 'transmitToClient', is_reply_correct)
    monkeypatch.setattr(node, 'send_ack_to_client', is_ack)
    node.process_action(req_obj, None)


def _comparison_reply(resp, req_obj):
    assert resp.result[f.IDENTIFIER.nm] == req_obj.identifier
    assert resp.result[f.REQ_ID.nm] == req_obj.reqId
    assert resp.result[TXN_TYPE] == VALIDATOR_INFO
    assert resp.result[DATA] is not None





import json

import pytest

from common.exceptions import LogicError
from indy_common.constants import RS_CONTENT, JSON_LD_CONTEXT_FIELD, ENDORSER
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.json_ld_context_handler import \
    JsonLdContextHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import context_request
from indy_node.test.rich_schema.templates import W3C_BASE_CONTEXT, W3C_EXAMPLE_V1_CONTEXT
from plenum.common.constants import TXN_TYPE, TRUSTEE
from plenum.common.exceptions import InvalidClientRequest


@pytest.fixture()
def context_handler(db_manager, write_auth_req_validator):
    return JsonLdContextHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def context_req(context_handler):
    req = context_request()
    add_to_idr(context_handler.database_manager.idr_cache, req.identifier, TRUSTEE)
    add_to_idr(context_handler.database_manager.idr_cache, req.endorser, ENDORSER)
    return req


def test_static_validation_context_no_context_field(context_handler, context_req):
    context_req.operation[RS_CONTENT] = json.dumps({"aaa": "2http:/..@#$"})
    with pytest.raises(InvalidClientRequest) as e:
        context_handler.static_validation(context_req)

    assert "must contain a @context field" in str(e.value)


def test_static_validation_context_fail_bad_uri(context_handler, context_req):
    context_req.operation[RS_CONTENT] = json.dumps({JSON_LD_CONTEXT_FIELD: "2http:/..@#$"})
    with pytest.raises(InvalidClientRequest) as e:
        context_handler.static_validation(context_req)

    assert "@context URI 2http:/..@#$ badly formed" in str(e.value)


def test_static_validation_fail_context_not_uri_or_array_or_object(context_handler, context_req):
    context_req.operation[RS_CONTENT] = json.dumps({JSON_LD_CONTEXT_FIELD: 52})
    with pytest.raises(InvalidClientRequest) as e:
        context_handler.static_validation(context_req)

    assert "'@context' value must be url, array, or object" in str(e.value)


def test_static_validation_pass_context_value_is_dict(context_handler, context_req):
    context = {
        "favoriteColor": "https://example.com/vocab#favoriteColor"
    }
    context_req.operation[RS_CONTENT] = json.dumps({JSON_LD_CONTEXT_FIELD: context})
    context_handler.static_validation(context_req)


def test_static_validation_pass_context_value_is_list_with_dict_and_uri(context_handler, context_req):
    context = [
        {
            "favoriteColor": "https://example.com/vocab#favoriteColor"
        },
        "https://www.w3.org/ns/odrl.jsonld"
    ]
    context_req.operation[RS_CONTENT] = json.dumps({JSON_LD_CONTEXT_FIELD: context})
    context_handler.static_validation(context_req)


def test_static_validation_pass_context_w3c_example_15(context_handler, context_req):
    context = {
        "@context": {
            "referenceNumber": "https://example.com/vocab#referenceNumber",
            "favoriteFood": "https://example.com/vocab#favoriteFood"
        }
    }
    context_req.operation[RS_CONTENT] = json.dumps(context)
    context_handler.static_validation(context_req)


def test_static_validation_fail_context_is_list_with_dict_and_bad_uri(context_handler, context_req):
    context = [
        {
            "favoriteColor": "https://example.com/vocab#favoriteColor"
        },
        "this is a bad uri"
    ]
    context_req.operation[RS_CONTENT] = json.dumps({JSON_LD_CONTEXT_FIELD: context})
    with pytest.raises(InvalidClientRequest) as e:
        context_handler.static_validation(context_req)

    assert "@context URI this is a bad uri badly formed" in str(e.value)


def test_static_validation_pass_context_w3c_base(context_handler, context_req):
    # Sample from specification: https://w3c.github.io/vc-data-model/#base-context
    # Actual file contents from: https://www.w3.org/2018/credentials/v1
    context_req.operation[RS_CONTENT] = json.dumps(W3C_BASE_CONTEXT)
    context_handler.static_validation(context_req)


def test_static_validation_pass_context_w3c_examples_v1(context_handler, context_req):
    # test for https://www.w3.org/2018/credentials/examples/v1
    context_req.operation[RS_CONTENT] = json.dumps(W3C_EXAMPLE_V1_CONTEXT)
    context_handler.static_validation(context_req)


def test_static_validation_fail_invalid_type(context_handler, context_req):
    context_req.operation[TXN_TYPE] = "201"
    with pytest.raises(LogicError):
        context_handler.static_validation(context_req)


def test_schema_dynamic_validation_passes(context_handler, context_req):
    context_handler.dynamic_validation(context_req, 0)



import pytest
from indy_common.types import ClientGetAttribOperation
from collections import OrderedDict
from plenum.common.messages.fields import ConstantField, IntegerField, LimitedLengthStringField, IdentifierField, Sha256HexField, TxnSeqNoField


EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("dest", IdentifierField),
    ("raw", LimitedLengthStringField),
    ('enc', LimitedLengthStringField),
    ('hash', Sha256HexField),
    ('timestamp', IntegerField),
    ('seqNo', TxnSeqNoField)
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientGetAttribOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientGetAttribOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)



from indy_common.constants import ENDORSER, NETWORK_MONITOR
from plenum.common.constants import TRUSTEE, STEWARD


OTHER_ROLE = "OtherRole"
OTHER_IDENTIFIER = "some_other_identifier"


IDENTIFIERS = {TRUSTEE: ["trustee_identifier", "trustee_identifier2", "trustee_identifier3", "trustee_identifier4"],
               STEWARD: ["steward_identifier", "steward_identifier2", "steward_identifier3", "steward_identifier4"],
               ENDORSER: ["endorser_identifier", "endorser_identifier2", "endorser_identifier3",
                          "endorser_identifier4"],
               NETWORK_MONITOR: ["network_monitor_identifier"],
               None: ["identity_owner_identifier", "identity_owner_identifier2", "identity_owner_identifier3",
                      "identity_owner_identifier4"],
               OTHER_ROLE: [OTHER_IDENTIFIER, "some_other_identifier2", "some_other_identifier3",
                            "some_other_identifier4"]}
